/*
 * Segment 2 (KEx),  SWE 2.Statischer Import Netz
 * Copyright (C) 2007 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weienfelser Strae 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.bwsin.konfig;

import org.geotools.feature.Feature;

import de.bsvrz.kex.bwsin.konfig.okstra.IllegalAttributeConversionException;
import de.bsvrz.kex.bwsin.konfig.okstra.Okstra;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraBetriebsKilometer;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraBlock;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraFeatures;

/**
 * Repr&auml;sentiert eine Betriebskilometer-Angabe.
 * 
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: BetriebsKilometerPunkt.java 20831 2009-12-17 16:14:48Z gieseler $
 * 
 */
public class BetriebsKilometerPunkt implements Cloneable {

	/**
	 * Die Feature-ID des Abschnittes/Astes auf dem der Betriebskilometer liegt.
	 */
	private String aoaFeatureID = null;

	/**
	 * Das zugeh&ouml;rige OKSTRA-Feature.
	 */
	private Feature feature = null;

	/**
	 * Stationierungswert.
	 */
	private int stationierung = 0;

	/**
	 * Kilometrierung.
	 */
	private double kilometrierung;

	/**
	 * Blocknummer.
	 */
	private int blocknummer;

	/**
	 * Offset auf dem Linienobjekt.
	 */
	private int offset;

	/**
	 * Erzeugt eine Betriebskilometer-Angabe.
	 * 
	 * @param aoaid
	 *            Feature-ID des Abschnittes/Astes auf dem der Betriebskilometer
	 *            liegt
	 * @param feature
	 *            das zugeh&ouml;rige OKSTRA-Feature.
	 * @throws ImportException
	 *             wenn der Betriebskilometer nicht angelegt werden konnte
	 */
	public BetriebsKilometerPunkt(final String aoaid, final Feature feature)
			throws ImportException {
		this.aoaFeatureID = aoaid;
		this.feature = feature;

		this.stationierung = Integer.parseInt(Okstra.getInstance()
				.getSpecialAttribute2(OkstraFeatures.BETRIEBSKILOMETER,
						feature.getID(), OkstraBetriebsKilometer.STATION));

		bestimmeKilometrierung();
		bestimmeBlocknummer();
	}

	/**
	 * Private Konstruktor.
	 */
	private BetriebsKilometerPunkt() {
	}

	/**
	 * Bestimmt die Blocknummer des Betriebskilometers.
	 * 
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	private void bestimmeBlocknummer() throws ImportException {
		try {
			String blockfeatureid = (String) feature
					.getAttribute(OkstraBetriebsKilometer.BLOCK);
			Feature block = Okstra.getFeatureReferenz(blockfeatureid);
			blocknummer = Okstra.getIntAttribute(block, OkstraBlock.BLOCK);
		} catch (IllegalAttributeConversionException e1) {
			throw new ImportException("Die Blocknummer des Features "
					+ feature.getID() + " kann nicht bestimmt werden");
		}
	}

	/**
	 * Bestimmt die Kilometrierungsangabe.
	 * 
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	private void bestimmeKilometrierung() throws ImportException {
		try {
			kilometrierung = Okstra.getInstance().getDoubleAttribute(feature,
					OkstraBetriebsKilometer.KILOMETRIERUNG);
		} catch (IllegalAttributeConversionException e1) {
			throw new ImportException("Die Kilometrierung des Features "
					+ feature.getID() + " kann nicht bestimmt werden");
		}
	}

	/**
	 * Gibt die Feature-ID des Abschnittes/Astes, auf dem der Betriebskilometer
	 * liegt, zur&uuml;ck.
	 * 
	 * @return Feature-ID Abschnitt/Ast
	 */
	public String getAoa() {
		return aoaFeatureID;
	}

	/**
	 * Gibt die Blocknummer zur&uuml;ck.
	 * 
	 * @return Blocknummer
	 */
	public int getBlocknummer() {
		return blocknummer;
	}

	/**
	 * Gibt den Datenkatalog-Wert f&uuml;r die Kilometrierung zur&uuml;ck. Im
	 * Datenkatalog wird der Betriebskilometer in m repr&auml;sentiert, der
	 * OKSTRA-Wert ist in km.
	 * 
	 * @return Kilometrierungswert f&&uml;r den Datenkatalog
	 */
	public int getDavWert() {
		return (int) (kilometrierung * 1000);
	}

	/**
	 * Gibt die Kilometrierung zur&uuml;ck.
	 * 
	 * @return Kilometrierung
	 */
	public double getKilometrierung() {
		return kilometrierung;
	}

	/**
	 * Gibt den Offset zur&uuml;ck.
	 * 
	 * @return Offset
	 */
	public int getOffset() {
		return offset;
	}

	/**
	 * Gibt die Kilometrierung zur&uuml;ck.
	 * 
	 * @return Kilometrierung
	 */
	public int getStationierung() {
		return stationierung;
	}

	/**
	 * Setzt den Offset.
	 * 
	 * @param offset
	 *            Offset des Punktes auf dem Linienelement.
	 */
	public void setOffset(final int offset) {
		this.offset = offset;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see java.lang.Object#clone()
	 */
	@Override
	public BetriebsKilometerPunkt clone() {
		BetriebsKilometerPunkt neu = new BetriebsKilometerPunkt();
		
		neu.aoaFeatureID = this.aoaFeatureID;
		neu.blocknummer = this.blocknummer;
		neu.feature = this.feature;
		neu.stationierung = this.stationierung;
		neu.kilometrierung = this.kilometrierung;

		return neu;
	}
}
