/*
 * Segment 2 (KEx),  SWE 2.Statischer Import Netz
 * Copyright (C) 2007 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weienfelser Strae 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.bwsin.konfig;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.geotools.feature.Feature;

import de.bsvrz.kex.bwsin.konfig.aktuell.AufteilungspunktAktuelleKonfiguration;
import de.bsvrz.kex.bwsin.konfig.kb.KonfigurationsObjekt;
import de.bsvrz.kex.bwsin.konfig.kb.SINASS;
import de.bsvrz.kex.bwsin.konfig.kb.SINSTS;
import de.bsvrz.kex.bwsin.konfig.lcl.LCL;
import de.bsvrz.kex.bwsin.konfig.okstra.Okstra;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraStrassenElement;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraStrassenElementAOA;
import de.bsvrz.kex.bwsin.vew.SINProperties;
import de.bsvrz.sys.funclib.bitctrl.geolib.WGS84Polygon;
import de.bsvrz.sys.funclib.bitctrl.geolib.WGS84Punkt;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Ein Strassenabschnitt repr&auml;sentiert eine Liste von zusammengefassten
 * Stra&szlig;enelementen.
 * 
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: StrassenAbschnitt.java 10402 2008-07-10 09:07:00Z gieseler $
 * 
 */
public class StrassenAbschnitt extends LinkedList<StrassenElement> {
	/**
	 * serialVersionUID.
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * DebugLogger fr Debug-Ausgaben.
	 */
	private static final Debug DEBUG = Debug.getLogger();

	/**
	 * Die Stra&szlig;enteilsegmente.
	 */
	private final ArrayList<KonfigurationsObjekt> teilsegmente = new ArrayList<KonfigurationsObjekt>();

	/**
	 * Der Start-Verbindungspunkt.
	 */
	protected String startVP = ""; //$NON-NLS-1$

	/**
	 * Der End-Verbindungspunkt.
	 */
	protected String endeVP = ""; //$NON-NLS-1$

	/**
	 * Flag, ob der Abschnitt bereits einem Stra&szlig;ensegment zugeordnet
	 * wurde.
	 */
	private boolean zugeordnet = false;

	/**
	 * Die Liste der zugeordneten ASS.
	 */
	private final List<SINASS> assListe = new ArrayList<SINASS>();

	/**
	 * Die Verkehrsrichtung in Bezug zur Stationierungsrichtung.
	 */
	private Verkehrsrichtung richtung = Verkehrsrichtung.RICHTUNG_UNBEKANNT;

	/**
	 * Liste der Teilabschnitte, asu denen der Abschnitt besteht.
	 */
	List<StrassenTeilAbschnitt> teilAbschnitte = new ArrayList<StrassenTeilAbschnitt>();

	/**
	 * Privater Konstruktor.
	 */
	private StrassenAbschnitt() {
		super();
	}

	/**
	 * Erzeugt einen Stra&szlig;enabschnitt fr ein Stra&szlig;enelement.
	 * 
	 * @param strassenelement
	 *            {@link StrassenElement}
	 */
	public StrassenAbschnitt(StrassenElement strassenelement) {
		add(strassenelement);
		startVP = strassenelement.getStartVP();
		endeVP = strassenelement.getEndeVP();
		// _richtung = strassenelement.getUnbenutzteRichtung();
		// _nachLocations.addAll(strassenelement.getNachLocationen());
	}

	/**
	 * F&uuml;gt ein Element zur Liste der auf diesem Strassenabschnitt
	 * liegenden ASS hinzu.
	 * 
	 * @param ass
	 *            {@link SINASS}
	 */
	public void addASS(SINASS ass) {
		this.assListe.add(ass);
	}

	/**
	 * F&uuml;gt ein Teilsegment hinzu.
	 * 
	 * @param teilsegment
	 *            Teilsegment
	 */
	public void addSTS(final KonfigurationsObjekt teilsegment) {
		teilsegmente.add(teilsegment);
	}

	/**
	 * F&uuml;gt ein Strassenelement am Anfang des Strassenabschnittes an.
	 * 
	 * @param strassenelement
	 *            anzuf&uuml;gendes Strassenelement
	 */
	public void anfuegenAnfang(StrassenElement strassenelement) {
		this.addFirst(strassenelement); // einfgen
		startVP = strassenelement.getStartVP();
	}

	/**
	 * F&uuml;gt ein Strassenelement am Ende des Strassenabschnittes an.
	 * 
	 * @param strassenelement
	 *            anzuf&uuml;gendes Strassenelement
	 */
	public void anfuegenEnde(StrassenElement strassenelement) {
		this.addLast(strassenelement);
		endeVP = strassenelement.getEndeVP();
	}

	/**
	 * Teilt den Str&szlig;enabschnitt an einer Liste von Aufteilungspunkten
	 * auf. Der Abschnitt wird nur geteilt, wenn der Punkt auf ihm liegt.
	 * 
	 * @param list
	 *            Liste von Aufteilungspunkten
	 * @return Liste von Aufteilungspunkten, die bei der Aufteilung des
	 *         Abschnittes benutzt wurden
	 */
	public List<AufteilungspunktAktuelleKonfiguration> aufteilenKoordinaten(
			List<AufteilungspunktAktuelleKonfiguration> list) {

		List<AufteilungspunktAktuelleKonfiguration> benutztePunkte = new ArrayList<AufteilungspunktAktuelleKonfiguration>();

		for (AufteilungspunktAktuelleKonfiguration ap : list) {
			for (int i = 0; i < teilAbschnitte.size(); i++) {
				if (teilabschnittAufteilen(i, ap)) {
					benutztePunkte.add(ap);
					break; // pro Punkt sollte nur eine Aufteilung erfolgen
				}
			}
		}

		return benutztePunkte;
	}

	/**
	 * Teilt den Str&szlig;enabschnitt bei einem Offset.
	 * 
	 * @param offset
	 *            Offset Liste von Aufteilungspunkten
	 * @return aufgeteilt ja/nein
	 */
	public boolean aufteilenOffset(float offset) {
		double aoffset = offset;
		for (int i = 0; i < teilAbschnitte.size(); i++) {
			if (teilAbschnitte.get(i).laenge() > offset) {
				teilabschnittAufteilen(i, offset);
				return true;
			}
			aoffset -= teilAbschnitte.get(i).laenge();
		}
		return false;
	}

	/**
	 * Berechnet den L&auml;ngenkorrekturfaktor f&uuml;r den Abschnitt als
	 * Verh&auml;ltnis von Polygonl&auml;nge zu ASB-Stationierungsl&auml;nge.
	 * 
	 * @param polygon
	 *            Polygonzug
	 * @param teilabschnitte
	 *            die Streckenteilabschnitte des Abschnittes
	 * @return L&auml;ngenkorrekturfaktor als Verh&auml;ltnis von
	 *         Polygonl&auml;nge zu ASB-Stationierungsl&auml;nge.
	 */
	private double berechneLaengenkorrektur(final WGS84Polygon polygon,
			final StreckenTeilabschnitte teilabschnitte) {
		// berechne ASB-Lnge ber alle Streckenteilabschnitte
		int laenge = 0;
		for (StreckenTeilAbschnitt stat : teilabschnitte) {
			laenge += stat.getStationierungEnde()
					- stat.getStationierungAnfang();
		}

		if (laenge != 0) {
			return polygon.laenge() / laenge;
		}

		return 1.0;
	}

	/**
	 * Berechnet die Stationierungsl&auml;nge einer Liste von
	 * ASB-Stationierungsangaben.
	 * 
	 * @param stationierungen
	 *            Liste von ASB-Stationierungsangaben
	 * @return Stationierungsl&auml;nge
	 */
	private int berechneStationierungsLaenge(
			List<AsbStationierung> stationierungen) {
		int laenge = 0;
		for (AsbStationierung asb : stationierungen) {
			laenge += asb.getStationierungEnde() - asb.getStationierungAnfang();
		}

		return laenge;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see java.util.LinkedList#clone()
	 */
	@Override
	public StrassenAbschnitt clone() {
		DEBUG.finest("Der Abschnitt '" + this + "' wird geclont"); //$NON-NLS-1$//$NON-NLS-2$
		StrassenAbschnitt neu = new StrassenAbschnitt();
		neu.addAll(this);
		neu.startVP = this.startVP;
		neu.endeVP = this.endeVP;
		// for(KonfigurationsObjekt ko : this._teilsegmente)
		// neu._teilsegmente.add(((SINSTS)ko).clone());
		for (KonfigurationsObjekt ko : this.teilsegmente) {
			neu.teilsegmente.add(ko);
		}
		neu.richtung = this.richtung;

		return neu;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see java.util.AbstractList#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (!super.equals(obj)) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final StrassenAbschnitt other = (StrassenAbschnitt) obj;
		if (endeVP == null) {
			if (other.endeVP != null) {
				return false;
			}
		} else if (!endeVP.equals(other.endeVP)) {
			return false;
		}
		if (this.get(0).getRichtung() != other.get(0).getRichtung()) {
			return false;
		}
		if (startVP == null) {
			if (other.startVP != null) {
				return false;
			}
		} else if (!startVP.equals(other.startVP)) {
			return false;
		}
		return true;
	}

	/**
	 * Erbt Features.
	 * 
	 * @param vererbteFeatures
	 *            Liste der bereits geerbten Features
	 * @param teilungspunkt
	 *            Teilungspunkt, von dem die Features geerbt werden sollen
	 */
	private void erbeFeatures(List<Feature> vererbteFeatures,
			Teilungspunkt teilungspunkt) {
		// erbe beginnende Features
		for (Feature f : teilungspunkt.getFeaturesBeginn()) {
			if (!vererbteFeatures.contains(f)) {
				vererbteFeatures.add(f);
			}
		}

		// loesche endende Features
		for (Feature f : teilungspunkt.getFeaturesEnde()) {
			if (vererbteFeatures.contains(f)) {
				vererbteFeatures.remove(f);
			}
		}
	}

	/**
	 * Gibt die Feature-ID des Verbindungspunktes zur&uuml;ck, an dem der
	 * Stra&szlig;enabschnitt beginnt.
	 * 
	 * @return Feature-ID des ersten Verbindungspunktes
	 */
	public String ersterVP() {
		return startVP;
	}

	/**
	 * Erzeugt die Teilabschnittes des Straenabschnitts durch Aufteilung an den
	 * Teilungspunkten der OKSTRA-Datenbasis.
	 * 
	 * @param okstra
	 *            Die Okstra-Datenbasis
	 * @param tp
	 *            Teilungspunkte
	 */
	public void erzeugeTeilAbschnitte(Okstra okstra, TeilungspunkteOkstra tp) {

		// Erzeuge die Teilabschnitte fr alle Straensegmente dieses
		// Abschnittes
		Iterator<StrassenElement> itse = this.iterator();
		while (itse.hasNext()) {
			StrassenElement se = itse.next();
			try {
				List<StrassenTeilAbschnitt> element = erzeugeTeilAbschnitte(se,
						tp);

				if (se.getRichtung() == Verkehrsrichtung.RICHTUNG_GEGEN_STATIONIERUNGSRICHTUNG) {
					DEBUG
							.fine("Die Verkehrsrichtung des Strassenelementes '" + se + "' wird umgekehrt"); //$NON-NLS-2$

					Collections.reverse(element);

					// Teilsegmente Koordinaten umkehren
					for (StrassenTeilAbschnitt s : element) {
						Collections.reverse(s.getKoordinaten());
					}
				}

				teilAbschnitte.addAll(element);
			} catch (ImportException ie) {
				DEBUG.error(ie.getMessage());
				continue;
			}
		}
	}

	/**
	 * Erzeugt eine Liste von Stra&szlig;enteilabschnitten durch Aufteilung des
	 * Stra&szlig;enabschnittes an den Aufteilungspunkten.
	 * 
	 * @param strassenelement
	 *            Stra&szlig;enelement
	 * @param teilungspunkte
	 *            Aufteilungspunkte
	 * @return Liste von Teilabschnitten
	 * @throws ImportException
	 *             bei Ausnahmen
	 * 
	 */
	private List<StrassenTeilAbschnitt> erzeugeTeilAbschnitte(
			StrassenElement strassenelement, TeilungspunkteOkstra teilungspunkte)
			throws ImportException {

		LinkedList<StrassenTeilAbschnitt> erzeugteTeilabschnitte = new LinkedList<StrassenTeilAbschnitt>();
		List<Feature> vererbteFeatures = new ArrayList<Feature>();

		List<AsbStationierung> stationierungen = new ArrayList<AsbStationierung>();

		DEBUG.fine("erzeuge TeilAbschnitte fr " + strassenelement.getID()); //$NON-NLS-1$

		WGS84Polygon apoly = Okstra
				.getFeatureKoordinatenAlsWGS84Polygon(strassenelement
						.getFeature());

		// die Streckenteilabschnitte des Streckenelementes bestimmen
		StreckenTeilabschnitte teilabschnitte;
		try {
			teilabschnitte = getTeilabschnitte(strassenelement.getFeature());
		} catch (ImportException e1) {
			DEBUG.warning("Das Streckenelement '" + strassenelement.getID() + //$NON-NLS-1$
					"') kann nicht aufgeteilt werden (" + //$NON-NLS-1$
					e1.getMessage() + ")"); //$NON-NLS-1$
			// es wird aber trotzdem ein Teilabschnitt erstellt
			teilabschnitte = null;
		}

		if (teilabschnitte != null) {
			// Laengenkorrektur berechenen
			double laengenkorrektur = berechneLaengenkorrektur(apoly,
					teilabschnitte);

			// ueber alle Streckenteilabschnitte des Streckenelementes
			for (StreckenTeilAbschnitt teilabschnitt : teilabschnitte) {
				AsbStationierung asbTeilabschnitt = teilabschnitt.clone();
				asbTeilabschnitt.setVerkehrsRichtung(strassenelement
						.getRichtung());

				List<Teilungspunkt> teilungspunkteabschnitt = null;
				try {
					teilungspunkteabschnitt = teilabschnitt
							.bestimmeTeilungspunkte(teilungspunkte);
				} catch (ImportException ie) {
					DEBUG.error(ie.getMessage());
					break;
				}

				if (teilungspunkteabschnitt.size() == 0) {
					DEBUG.finer("keine Aufteilungspunkte fr AoA: " + //$NON-NLS-1$
							teilabschnitt.getAOAId());
					continue;
				}

				int anfangsOffset = teilabschnitt.getStationierungAnfang();
				int endeOffset = teilabschnitt.getStationierungEnde();

				if (anfangsOffset > endeOffset) {
					DEBUG
							.error("Die Anfangsstationierung ist grer als die Endstationierung (" + //$NON-NLS-1$
									teilabschnitt.streckenFeatureID + ")"); //$NON-NLS-1$
					continue;
				}

				int anzahlTeilabschnitte = 0;

				for (Teilungspunkt teilungspunkt : teilungspunkteabschnitt) {
					// Stationierung des Punktes
					int stationierung = teilungspunkt.getStationierung();

					// erbeFeatures(vererbteFeatures, teilungspunkt);

					if (stationierung <= anfangsOffset) {
						// der Teilungspunkt liegt vor oder am Anfang der
						// Strecke -> ererbe alle Features
						erbeFeatures(vererbteFeatures, teilungspunkt);
						continue;
					}

					if (stationierung > endeOffset) {
						// der Teilungspunkt liegt hinter der Strecke, die
						// Teilungspunkte sind geordnet --> Ende
						break;
					}

					teilungspunkt.setUsed(); // fr interne Tests

					// berechne abzuschneidendes Polygonstueck
					int stationierungAbschneiden = berechneStationierungsLaenge(stationierungen)
							+ stationierung - anfangsOffset;
					double ablaenge = Math.round(stationierungAbschneiden
							* laengenkorrektur * 1000.0) / 1000.0;

					double plaenge = apoly.laenge();

					// korrigiere numerischen Fehler
					if (ablaenge > plaenge) {
						// erlaube 1m
						if (Math.abs(ablaenge - plaenge) > 1.0) {
							throw new IllegalArgumentException(
									"Lnge zu gro: ab: " + ablaenge //$NON-NLS-1$
											+ " polygon: " + plaenge); //$NON-NLS-1$
						}
						DEBUG
								.finer("abzuschneiden korrigiert von " + ablaenge + " auf " + plaenge); //$NON-NLS-1$ //$NON-NLS-2$
						ablaenge = plaenge;
					}

					WGS84Polygon teilpolygon;
					try {
						teilpolygon = apoly.anfangAbschneiden(ablaenge);
					} catch (IllegalStateException ise) {
						throw new ImportException(
								"Aufteilungsfehler Polygon (" + ise.getMessage() + ")"); //$NON-NLS-1$ //$NON-NLS-2$
					}

					StrassenTeilAbschnitt neuerTeilabschnitt = null;

					if (teilpolygon != null) {
						if (erzeugteTeilabschnitte.size() == 0) {
							// erster Teilabschnitt erbt alle Eigenschaften
							// erbeFeatures(vererbteFeatures, teilungspunkt);
							neuerTeilabschnitt = new StrassenTeilAbschnitt(
									teilpolygon, strassenelement,
									vererbteFeatures);
						} else {
							neuerTeilabschnitt = new StrassenTeilAbschnitt(
									teilpolygon, strassenelement,
									erzeugteTeilabschnitte.getLast()
											.getEigenschaften());
							neuerTeilabschnitt
									.aktualisiereFeatures(teilungspunkt);
						}

						// Liste der bergangenen Stationierungen hinzufgen
						neuerTeilabschnitt
								.setAsbStationierungen(stationierungen);
						stationierungen.clear();

						// Stationierung dieses Teilabschnittes
						AsbStationierung asbNeuerTeilabschnitt = asbTeilabschnitt
								.clone();
						asbNeuerTeilabschnitt.stationierungAnfang = anfangsOffset;
						asbNeuerTeilabschnitt.stationierungEnde = teilungspunkt
								.getStationierung();
						asbNeuerTeilabschnitt.verkehrsRichtung = strassenelement
								.getRichtung();
						neuerTeilabschnitt
								.setAsbStationierung(asbNeuerTeilabschnitt);

						anfangsOffset = stationierung;

						erzeugteTeilabschnitte.add(neuerTeilabschnitt);
						anzahlTeilabschnitte++;
					}
				}

				if ((anfangsOffset != endeOffset)
						|| (anzahlTeilabschnitte == 0)) {
					asbTeilabschnitt.stationierungAnfang = anfangsOffset;
					stationierungen.add(asbTeilabschnitt);
				}
			}
		}

		// Restpolygon
		if (apoly.laenge() > 0) {
			List<Feature> fl = erzeugteTeilabschnitte.size() > 0 ? erzeugteTeilabschnitte
					.getLast().getEigenschaften()
					: vererbteFeatures;

			StrassenTeilAbschnitt nsta = new StrassenTeilAbschnitt(apoly,
					strassenelement, fl);
			// setze reale Lnge
			// nsta.set_laenge(end_offset-anf_offset);
			// nsta.set_asb_offset(anf_offset);
			nsta.setAsbStationierungen(stationierungen);
			stationierungen.clear();
			erzeugteTeilabschnitte.add(nsta);
		}

		if (erzeugteTeilabschnitte.size() == 0) {
			// keine Aufteilungspunkte
			DEBUG.fine("Straenabschnitt nicht aufgeteilt"); //$NON-NLS-1$
			StrassenTeilAbschnitt nsta = new StrassenTeilAbschnitt(apoly,
					strassenelement, vererbteFeatures);
			// setze reale Lnge
			// nsta.set_laenge(end_offset-anf_offset);
			// nsta.set_asb_offset(anf_offset);
			nsta.setAsbStationierungen(stationierungen);
			erzeugteTeilabschnitte.add(nsta);
		} else {
			if (stationierungen.size() > 0) {
				erzeugteTeilabschnitte.getLast().setAsbStationierungen(
						stationierungen);
			}
		}

		DEBUG.fine("erzeugte TeilAbschnitte: " + erzeugteTeilabschnitte.size()); //$NON-NLS-1$

		return erzeugteTeilabschnitte;
	}

	/**
	 * Gibt die Liste der auf diesem Strassenabschnitt liegenden ASS
	 * zur&uuml;ck.
	 * 
	 * @return Liste von {@link SINASS}
	 */
	public List<SINASS> getASS() {
		return assListe;
	}

	/**
	 * Gibt das auf diesem Strassenabschnitt liegende ASS zur&uuml;ck, welches
	 * sich auf eine Location zubewegt.
	 * 
	 * @param location
	 *            Location
	 * @return {@link KonfigurationsObjekt}
	 */
	public KonfigurationsObjekt getASSnach(Location location) {
		for (SINASS ass : assListe) {
			if (ass.getLocationNach().equals(location)) {
				return ass;
			}
		}

		return null;
	}

	/**
	 * Gibt das auf diesem Strassenabschnitt liegende ASS zur&uuml;ck, welches
	 * sich von einer Location wegbewegt.
	 * 
	 * @param location
	 *            Location
	 * @return {@link KonfigurationsObjekt}
	 */
	public KonfigurationsObjekt getASSvon(Location location) {
		for (SINASS ass : assListe) {
			if (ass.getLocationVon().equals(location)) {
				return ass;
			}
		}

		return null;
	}

	/**
	 * Gibt die L&auml;nge des Abschnittes zur&uuml;ck.
	 * 
	 * @return L&auml;nge
	 */
	public double getLaenge() {
		return getPolygon().laenge();
	}

	/**
	 * Gibt die Liste der Locationen zur&uuml;ck, auf die sich der
	 * Stra&szlig;en- abschnitt zubewegt.
	 * 
	 * @return Liste von {@link Location}
	 */
	public Location getLocationNach() {
		return getFirst().getNachLocation();
	}

	/**
	 * Bestimmt den Anfangs-Locationcode des Strassenabschnittes.
	 * 
	 * @param nach
	 *            Location, auf die sich der Abschnitt zubewegt
	 * @return Locationcode Anfangs-Location
	 * @throws ImportException
	 *             wenn keine Vorg&auml;ngerlocation bestimmt werden kann
	 */
	public Location getLocationVon(Location nach) throws ImportException {
		String spalte = nach.getLocationRichtung() == Location.LOCATION_RICHTUNG_POSITIV ? LCL.SPALTE_NACHFOLGER
				: LCL.SPALTE_VORGAENGER;
		List<String> lcl = LCL.getInstance().getLocationCodes(spalte,
				nach.getLocationCode());

		if (lcl.size() == 0) {
			throw new ImportException(
					"Die Anfangs-Location zur Location " + //$NON-NLS-1$
							nach
							+ " kann nicht bestimmt werden, da in der LCL kein Vorgnger gefunden wurde."); //$NON-NLS-1$
		}

		if (lcl.size() > 1) {
			throw new ImportException(
					"Die Anfangs-Location zur Location " + //$NON-NLS-1$
							nach
							+ " kann nicht bestimmt werden, da mehrere Vorgnger-Location-Codes " + //$NON-NLS-1$
							" exitieren."); //$NON-NLS-1$
		}

		return new Location(lcl.get(0), nach.getLocationRichtung());
	}

	/**
	 * Gibt die Verkehrsrichtung des Stra&szlig;enabschnittes zur&uuml;ck.
	 * 
	 * @return {@link Verkehrsrichtung}
	 */
	public Verkehrsrichtung getOkstraRichtung() {
		return getFirst().getRichtung();
	}

	/**
	 * Gibt den Polygonzug des Abschnittes zur&uuml;ck.
	 * 
	 * @return Polygonzug
	 */
	public WGS84Polygon getPolygon() {
		List<WGS84Punkt> punktliste = new ArrayList<WGS84Punkt>();

		for (StrassenTeilAbschnitt sta : teilAbschnitte) {
			punktliste.addAll(sta.getKoordinaten());
		}

		return new WGS84Polygon(punktliste);
	}

	/**
	 * Gibt die Verkehrsrichtung des Stra&szlig;enabschnittes zur&uuml;ck.
	 * 
	 * @return {@link Verkehrsrichtung}
	 */
	public Verkehrsrichtung getRichtung() {
		return richtung;
	}

	/**
	 * Gibt die Liste der Stra&szlig;enteilsegmente zur&uuml;ck.
	 * 
	 * @return Liste von {@link KonfigurationsObjekt}
	 */
	public List<KonfigurationsObjekt> getSTS() {
		return teilsegmente;
	}

	/**
	 * Gibt die Liste der Teilabschnitte zur&uuml;ck, aus denen der Abschnitt
	 * gebildet wird.
	 * 
	 * @return Liste von Teilabschnitten
	 */
	public List<StrassenTeilAbschnitt> getTeilabschnitte() {
		return teilAbschnitte;
	}

	/**
	 * Gibt die Streckenteilabschnitte der OKSTRA-Datensasis zu einem
	 * Stra&szlig;enelement zur&uuml;ck.
	 * 
	 * @param strassenelement
	 *            Stra&szlig;enelement
	 * @return Streckenteilabschnitte zum Stra&szlig;enelement
	 * @throws ImportException
	 *             wenn die Streckenteilabschnitte nicht bestimmt werden
	 *             k&ouml;nnen
	 */
	private StreckenTeilabschnitte getTeilabschnitte(Feature strassenelement)
			throws ImportException {
		try {
			String aufAOAId = (String) strassenelement
					.getAttribute(OkstraStrassenElement.STAT_AOA);
			if ((aufAOAId == null) || aufAOAId.length() == 0) {
				throw new ImportException(
						"Das Strassenelement " + strassenelement.getID() + //$NON-NLS-1$
								" besitzt keine AoA-Stationierungsangabe"); //$NON-NLS-1$
			}
			Feature aufAOA = Okstra.getFeatureReferenz(aufAOAId);
			if (aufAOA == null) {
				throw new ImportException(
						"Das referenzierte Feature " + aufAOAId + //$NON-NLS-1$
								" existiert nicht"); //$NON-NLS-1$
			}
			String strecke = Okstra.getRefFeatureId((String) aufAOA
					.getAttribute(OkstraStrassenElementAOA.STRECKE));
			return new StreckenTeilabschnitte(strecke);
		} catch (IllegalArgumentException ie) {
			throw new ImportException(
					"Die Streckenreferenz des Strassenelementes " //$NON-NLS-1$
							+ strassenelement
							+ " kann nicht bestimmt werden (" + //$NON-NLS-1$
							ie.getMessage() + ")"); //$NON-NLS-1$
		}
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see java.util.AbstractList#hashCode()
	 */
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = super.hashCode();
		result = prime * result + ((endeVP == null) ? 0 : endeVP.hashCode());
		result = prime * result + this.get(0).getRichtung().ordinal();
		result = prime * result + ((startVP == null) ? 0 : startVP.hashCode());
		return result;
	}

	/**
	 * Test, ob der Straenabschnitt zu einem &auml;ueren Straensegment
	 * geh&ouml;rt.
	 * 
	 * @return true, wenn der Stra&szlig;enabschnitt zu einem &auml;ueren
	 *         Stra&szlig;ensegment geh&ouml;rt, ansonsten false
	 */
	public boolean istASS() {
		return get(0).istASS();
	}

	/**
	 * Test, ob der Straenabschnitt der Beginn eines inneren Straensegmentes
	 * ist.
	 * 
	 * @param netz
	 *            Importnetz
	 * 
	 * @return true, wenn der Stra&szlig;enabschnitt der Beginn eines inneren
	 *         Straensegmentes ist, ansonsten false
	 */
	public boolean istISSAnfang(final ImportNetz netz) {
		if (istZugeordnet()) {
			return false;
		}

		if (istSegmentAnfang(netz)) {
			return true;
		}

		return false;
	}

	/**
	 * Test, ob der Stra&szlig;enabschnitt der Anfang eines Segmentes ist, d.h.
	 * es gibt keinen Vorg&auml;ngerabschnitt
	 * 
	 * @param netz
	 *            das Importnetz
	 * @return true, wenn der Stra&szlig;enabschnitt der Anfang eines Segmentes
	 *         ist, sonst false
	 */
	private boolean istSegmentAnfang(ImportNetz netz) {
		StrassenAbschnitt test = netz.getVorgaenger(this);

		if (test != null) {
			DEBUG.finest("kein Anfang: " + this);
		}

		return test == null;
	}

	/**
	 * Test, ob dieser {@link StrassenAbschnitt} bereits einem Stra&szlig;en-
	 * segment zugeordnet wurde.
	 * 
	 * @return true, wenn bereits zugeordnet, sonst false
	 */
	public boolean istZugeordnet() {
		return zugeordnet;
	}

	/**
	 * Test, ob der Straenabschnitt mit dem Stra&szlig;enelement verbunden
	 * werden kann. Folgende Bedingungen m&uuml;ssen dazu erf&uuml;llt sein: -
	 * der Strassenabschnitt muss einem &auml;u&szlig;eren Stra&szlig;ensegment
	 * zugeordnet sein - die Verkehrsrichtung muss &uuml;bereinstimmen - die
	 * TMC-Locationen m&uuml;ssen &uuml;bereinstimmen
	 * 
	 * @param strelement
	 *            Stra&szlig;enelement
	 * @return true, wenn o.g. Bedingungen erf&uuml;llt sind, sonst false
	 */
	public boolean kannVerbundenWerden(StrassenElement strelement) {
		if (this.istASS() && strelement.istASS()) {
			// Richtung testen
			if (richtung == Verkehrsrichtung.RICHTUNG_BEIDE_RICHTUNGEN) {
				DEBUG.error("Abschnitt in beiden Richtungen");
				return false;
			}

			return getFirst().ueberpruefeLocationen(strelement);
		}

		return false;
	}

	/**
	 * Gibt eine Kurzbeschreibung des Abschnittes zur&uuml;ck.
	 * 
	 * @return textuelle Kurzbeschreibung
	 */
	public String kurzInfo() {
		String ses = get(0).getID();
		String see = get(this.size() - 1).getID();
		if (ses.equals(see)) {
			return "(" + ses.substring(ses.indexOf('.') + 1) + ")"; //$NON-NLS-1$ //$NON-NLS-2$
		}

		return "(" + ses.substring(ses.indexOf('.') + 1) + //$NON-NLS-1$
				"," + see.substring(see.indexOf('.') + 1) + ")"; //$NON-NLS-1$//$NON-NLS-2$
	}

	/**
	 * Gibt die Feature-ID des Verbindungspunktes zur&uuml;ck, an dem der
	 * Stra&szlig;enabschnitt endet.
	 * 
	 * @return Feature-ID des letzten Verbindungspunktes
	 */
	public String letzterVP() {
		return endeVP;
	}

	/**
	 * Setzt die Verkehrsrichtung des Stra&szlig;enabschnittes.
	 * 
	 * @param richtung
	 *            Verkehrsrichtung
	 */
	public void setRichtung(final Verkehrsrichtung richtung) {
		this.richtung = richtung;
	}

	/**
	 * Setzt die Verkehrsrichtung des Stra&szlig;enabschnittes.
	 */
	public void setRichtungSTS() {
		for (KonfigurationsObjekt sts : teilsegmente) {
			((SINSTS) sts).setRichtung(richtung);
		}
	}

	/**
	 * Setzt das Flag, das diesen {@link StrassenAbschnitt} als einem
	 * Stra&szlig;en- segment zugeordnet markiert.
	 */
	public void setZugeordnet() {
		zugeordnet = true;
	}

	/**
	 * Teilt einen Teilabschnitt an einer Koordinate auf, wenn die Koordinate
	 * auf dem Teilabschnitt liegt.
	 * 
	 * @param index
	 *            Index des Teilabschnittes in der Liste der Teilabschnitte
	 * @param ap
	 *            Aufteilungspunkt
	 * @return true, wenn der Abschnitt geteilt wurde, sonst false
	 */
	private boolean teilabschnittAufteilen(final int index,
			final AufteilungspunktAktuelleKonfiguration ap) {

		double maxabweichung = SINProperties.getInstance()
				.getMaximaleAbweichungMeter();
		WGS84Punkt punkt = new WGS84Punkt(ap.getKoordinaten());

		StrassenTeilAbschnitt teilabschnitt = teilAbschnitte.get(index);

		// _debug.info("bearbeite Teilabschnitt " + index + " ("
		// + teilabschnitt.getKoordinaten().size() + " Koordinaten)");

		if (teilabschnitt.liegtAufPolygon(punkt, maxabweichung)) {
			// Test, ob der Punkt am Ende des Polygons liegt
			WGS84Punkt teilungspunkt = teilabschnitt.bildPunkt(punkt);
			if (teilabschnitt.istAnfangsOderEndPunkt(punkt, maxabweichung)) {
				DEBUG.finer("Aufteilung durch " + ap.getSystemObjekt().getPid()
						+ " wird am Ende des Polygones ignoriert");
				return true;
			}

			DEBUG.fine("" + this + " wird durch "
					+ ap.getSystemObjekt().getPid() + " geteilt");
			List<StrassenTeilAbschnitt> neueabschnitte = teilabschnitt
					.teilen(teilungspunkt);
			teilAbschnitte.remove(index);
			teilAbschnitte.addAll(index, neueabschnitte);
			return true;
		}

		return false;
	}

	/**
	 * Teilt einen Teilabschnitt an einem Offset auf.
	 * 
	 * @param index
	 *            Index des Teilabschnittes in der Liste der Teilabschnitte
	 * @param offset
	 *            Offset, bei dem geteilt werden soll
	 * @return true, wenn der Abschnitt geteilt wurde, sonst false
	 */
	private boolean teilabschnittAufteilen(int index, double offset) {

		StrassenTeilAbschnitt teilabschnitt = teilAbschnitte.get(index);

		WGS84Punkt teilungspunkt = teilabschnitt.bildPunkt(offset);
		if (teilabschnitt.istAnfangsOderEndPunkt(teilungspunkt)) {
			return true;
		}

		List<StrassenTeilAbschnitt> neueabschnitte = teilabschnitt
				.teilen(teilungspunkt);
		teilAbschnitte.remove(index);
		teilAbschnitte.addAll(index, neueabschnitte);

		return true;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see java.util.AbstractCollection#toString()
	 */
	@Override
	public String toString() {
		String ret = null;

		if (this.size() > 1) {
			ret = "Strassenabschnitt[" + getFirst() + "..." + getLast() + "]"; //$NON-NLS-1$
		} else {
			ret = "Strassenabschnitt[" + getFirst() + "]"; //$NON-NLS-1$
		}

		return ret;
	}
}
