/*
 * Segment 2 (KEx),  SWE 2.Statischer Import Netz
 * Copyright (C) 2007 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weienfelser Strae 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.bwsin.konfig;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.geotools.feature.Feature;

import de.bsvrz.kex.bwsin.konfig.okstra.IllegalAttributeConversionException;
import de.bsvrz.kex.bwsin.konfig.okstra.Okstra;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraFNeigung;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraFeature;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraFeatures;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Repr&auml;sentiert die Aufteilungspunkte f&uuml;r die Stra&szlig;enelemente
 * aus der OKSTRA-Datenbank.
 * <p>
 * Der Zugriff auf die Aufteilungspunkte erfolgt &uuml;ber den Abschnitt/Ast.
 * </p>
 * 
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: TeilungspunkteOkstra.java 10402 2008-07-10 09:07:00Z gieseler $
 * 
 */
public class TeilungspunkteOkstra {

	/**
	 * DebugLogger fr Debug-Ausgaben.
	 */
	private final Debug debug = Debug.getLogger();

	/**
	 * Zuordnungstabelle.
	 */
	private final Map<String, HashMap<Integer, Teilungspunkt>> aoAMap = new HashMap<String, HashMap<Integer, Teilungspunkt>>();

	/**
	 * F&uuml;gt einen Punkt zur Zuordnungstabelle hinzu.
	 * <p>
	 * Wenn bereits ein Punkt existiert, werden die Features des existierenden
	 * Punktes aktualisiert.
	 * </p>
	 * 
	 * @param punkt
	 *            Teilungspunkt
	 * @return 1, wenn ein neuer Punkt hinzugef&uuml;gt wurde, sonst 0
	 */
	private int addPunkt(final Teilungspunkt punkt) {
		HashMap<Integer, Teilungspunkt> pmap = aoAMap.get(punkt.getAoa());
		int neu = 0;

		if (aoAMap.get(punkt.getAoa()) == null) {
			pmap = new HashMap<Integer, Teilungspunkt>();
			aoAMap.put(punkt.getAoa(), pmap);
		}

		Teilungspunkt tfp = pmap.get(new Integer(punkt.getStationierung()));
		if (tfp == null) {
			pmap.put(new Integer(punkt.getStationierung()), punkt);
			tfp = punkt;
			neu = 1;
		} else {
			tfp.addFeaturesBeginn(punkt.getFeaturesBeginn(), true);
			tfp.addFeaturesEnde(punkt.getFeaturesEnde());
		}

		return neu;
	}

	/**
	 * Gibt alle Teilungspunkte aus, die nicht referenziert wurden.
	 */
	public void checkUsed() {
		for (HashMap<Integer, Teilungspunkt> h : aoAMap.values()) {
			for (Teilungspunkt l : h.values()) {
				if (l.getUsed() == 0) {
					System.out
							.println("nicht zugeordneter Teilungspunkt: " + l.getGrund()); //$NON-NLS-1$
				}
			}
		}
	}

	/**
	 * Liefert eine Liste von Aufteilungspunkten, die auf einem Abschnitt oder
	 * Ast liegen.
	 * <p>
	 * Die Liste ist nach Stationierung aufsteigend geordnet.
	 * </p>
	 * 
	 * @param aoa
	 *            Abschnitt oder Ast
	 * @return Liste von Teilungspunkten auf dieser AoA, aufsteigend nach
	 *         Stationierung sortiert
	 */
	public List<Teilungspunkt> getAoaPunkte(final String aoa) {
		List<Teilungspunkt> rlist = new ArrayList<Teilungspunkt>();
		if (aoAMap.get(aoa) != null) {
			rlist.addAll(aoAMap.get(aoa).values());

			Collections.sort(rlist, new Comparator<Teilungspunkt>() {
				public int compare(final Teilungspunkt o1,
						final Teilungspunkt o2) {
					return new Integer(o1.getStationierung())
							.compareTo(new Integer(o2.getStationierung()));
				}
			});
		}

		return rlist;
	}

	/**
	 * Initialisiert die Aufteilungspunkte aus einem OKSTRA-Feature.
	 * 
	 * @param okstradb
	 *            OKSTRA-Datenbasis
	 * @param of
	 *            Feature der OKSTRA-Datenbasis
	 * @return Anzahl der erzeugten Aufteilungspunkte
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	private int initFeature(final Okstra okstradb, OkstraFeature of)
			throws ImportException {
		int added = 0;
		Map<String, Feature> fmap;
		Feature feature;

		try {
			debug.info("Aufteilungspunkte '" + of.getFeatureName() + //$NON-NLS-1$
					"' werden initialisiert."); //$NON-NLS-1$

			fmap = okstradb.getFeatureMap(of.getFeatureOkstraName());

			for (Map.Entry<String, Feature> e : fmap.entrySet()) {
				feature = e.getValue();

				final String streckenid = Okstra
						.getRefFeatureId((String) feature
								.getAttribute("hat_Strecke")); //$NON-NLS-1$
				StreckenTeilabschnitte teilabschnitte = new StreckenTeilabschnitte(
						streckenid);

				for (final StreckenTeilAbschnitt teilabschnitt : teilabschnitte) {
					// Teilungspunkt Anfang
					added += addPunkt(new Teilungspunkt(teilabschnitt, feature,
							null));

					// Teilungspunkt Ende
					added += addPunkt(new Teilungspunkt(teilabschnitt, null,
							feature));
				}
			}

			debug.fine("" + added + " neue(r) Teilungspunkt(e)"); //$NON-NLS-1$ //$NON-NLS-2$

		} catch (final IllegalArgumentException e1) {
			debug.error(e1.getMessage());
		} catch (ImportException e) {
			throw new ImportException("OKSTRA-Fehler: Feature: "
					+ of.getFeatureOkstraName() + ": " + e.getMessage());
		}

		return added;
	}

	/**
	 * Initialisierungsfunktion f&uuml;r die Teilungspunkte.
	 * 
	 * @param okstradb
	 *            die Okstra-Datenbasis
	 * @return Anzahl der erzeugten Teilungspunkte
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	public int initialisiere(final Okstra okstradb) throws ImportException {
		int added = 0;

		debug.info("Aufteilungspunkte werden initialisiert."); //$NON-NLS-1$

		added += initLaengsneigung(okstradb);

		added += initFeature(okstradb, OkstraFeatures.ANZAHLFAHRSTREIFEN);
		added += initFeature(okstradb, OkstraFeatures.GESCHWINDIGKEIT);
		added += initFeature(okstradb, OkstraFeatures.UEBERHOLVERBOT);

		debug.info("" + added + " Teilungspunkt(e) erzeugt"); //$NON-NLS-1$//$NON-NLS-2$

		return added;
	}

	/**
	 * Initialisierung der Aufteilungspunkte 'Laengsneigung'.
	 * <p>
	 * Es werden nur die Features mit einer signifikanten &Auml;nderung der
	 * L&auml;ngsneigung benutzt. Als signifikant wird die &Auml;nderung des
	 * (gerundeten) ganzzahligen Wertes der Neigung angenommen, da im DatK nur
	 * ganzzahlige Fahrbahnneigungen angegeben werden k&ouml;nnen.
	 * </p>
	 * 
	 * @param okstradb
	 *            OKSTRA-Datenbasis
	 * @return Anzahl der erzeugten Aufteilungspunkte
	 */
	private int initLaengsneigung(Okstra okstradb) {
		int added = 0;

		try {
			Map<String, Feature> m = okstradb
					.getFeatureMap(OkstraFeatures.FAHRBAHNNEIGUNG
							.getFeatureOkstraName());

			debug
					.fine("Aufteilungspunkte '" + OkstraFeatures.FAHRBAHNNEIGUNG.getFeatureName() + //$NON-NLS-1$
							"' werden initialisiert."); //$NON-NLS-1$

			debug.fine("" + OkstraFeatures.FAHRBAHNNEIGUNG.getFeatureName() + //$NON-NLS-1$
					": " + m.size() + " Feature(s)"); //$NON-NLS-1$ //$NON-NLS-2$

			// zuerst alle Lngsneigungen nach AoA sortieren
			// Map mit Key AoA erstellen
			Map<String, ArrayList<Teilungspunkt>> aoaneigmap = new HashMap<String, ArrayList<Teilungspunkt>>();

			for (Map.Entry<String, Feature> e : m.entrySet()) {
				// ueber alle Streckenteilabschnitte
				Feature ln = e.getValue();

				String streckenid = Okstra.getRefFeatureId((String) ln
						.getAttribute(OkstraFNeigung.STRECKE));
				StreckenTeilabschnitte teilabschnitte = new StreckenTeilabschnitte(
						streckenid);
				for (StreckenTeilAbschnitt teilabschnitt : teilabschnitte) {
					Teilungspunkt ntp = new Teilungspunkt(teilabschnitt, ln,
							null);
					if (aoaneigmap.get(teilabschnitt.getAOAId()) != null) {
						aoaneigmap.get(teilabschnitt.getAOAId()).add(ntp);
					} else {
						ArrayList<Teilungspunkt> ar = new ArrayList<Teilungspunkt>();
						ar.add(ntp);
						aoaneigmap.put(teilabschnitt.getAOAId(), ar);
					}
				}
			}

			// AoA Listen nach Stationierung aufsteigend ordnen
			for (Map.Entry<String, ArrayList<Teilungspunkt>> l : aoaneigmap
					.entrySet()) {
				Collections.sort(l.getValue(), new Comparator<Teilungspunkt>() {
					public int compare(final Teilungspunkt fn1,
							final Teilungspunkt fn2) {
						int stat1 = fn1.getStationierung();
						int stat2 = fn2.getStationierung();
						return new Integer(stat1).compareTo(new Integer(stat2));
					}
				});
			}

			// Features filtern und Ergebnisliste erzeugen
			for (Map.Entry<String, ArrayList<Teilungspunkt>> e : aoaneigmap
					.entrySet()) {
				int letztesteigung = -1000;

				for (Teilungspunkt tp : e.getValue()) {
					Feature lf = tp.getFeaturesBeginn().get(0);
					int steigung;
					try {
						// gleich als gerundetes int anfordern, da DAV sowieso
						// nur
						// int akzeptiert
						steigung = Okstra.getIntAttribute(lf,
								OkstraFNeigung.STEIGUNG);
					} catch (IllegalAttributeConversionException e1) {
						debug.error(e1.getMessage());
						continue;
					}

					if (steigung == letztesteigung) {
						debug
								.finest("Feature '" + OkstraFeatures.FAHRBAHNNEIGUNG.getFeatureName() //$NON-NLS-1$
										+ "' ID: " + lf.getID() + " wird ignoriert"); //$NON-NLS-1$//$NON-NLS-2$
						continue;
					}

					letztesteigung = steigung;

					// retlist.add(lf);
					addPunkt(tp);
					added++;
				}
			}
		} catch (ImportException e) {
			debug.error(e.getMessage());
		}
		debug.fine("" + OkstraFeatures.FAHRBAHNNEIGUNG.getFeatureName() + //$NON-NLS-1$
				": " + added + " gefilterte Feature(s)"); //$NON-NLS-1$//$NON-NLS-2$

		return added;
	}

}
