/*
 * Segment 2 (KEx),  SWE 2.Statischer Import Netz
 * Copyright (C) 2007 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weienfelser Strae 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.bwsin.konfig;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Properties;

import org.junit.Test;

import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.dav.daf.main.ClientDavParameters;
import de.bsvrz.dav.daf.main.CommunicationError;
import de.bsvrz.dav.daf.main.ConnectionException;
import de.bsvrz.dav.daf.main.InconsistentLoginException;
import de.bsvrz.dav.daf.main.MissingParameterException;
import de.bsvrz.kex.bwsin.konfig.aktuell.SINAktuelleKonfiguration;
import de.bsvrz.kex.bwsin.konfig.kb.KonfigurationsObjekt;
import de.bsvrz.kex.bwsin.konfig.kb.SINASS;
import de.bsvrz.kex.bwsin.konfig.kb.SINKonfigurationsBereich;
import de.bsvrz.kex.bwsin.konfig.kb.SINSTS;
import de.bsvrz.kex.bwsin.konfig.kb.SINStrassenSegment;
import de.bsvrz.kex.bwsin.konfig.lcl.LCL;
import de.bsvrz.kex.bwsin.konfig.okstra.Okstra;
import de.bsvrz.sys.funclib.bitctrl.geolib.WGS84Punkt;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Testfall Import. Neben dem eigentlichen Import werden zus&auml;tzliche
 * Informationen erzeugt.
 * 
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: TestImport.java 10285 2008-07-07 08:30:01Z gieseler $
 * 
 */
public class TestImport {

	private final String kbPid = "kb.verkehrStrassenSegmenteBw";

	private Okstra odb;
	private ClientDavConnection dav;
	private SINKonfigurationsBereich kb;
	private File _kbfile;

	private final String asbzusatzinfodatei = "ZusatzinfosASB.txt";
	private final String zusatzinfodatei = "Zusatzinfos.txt";

	/**
	 * Erzeugt eine Liste aller &Auml;u&szlig;sseren Stra&szlig;ensegmente des
	 * neu erzeugten Konfigurationsbereiches in einer Datei. Die Liste kann zur
	 * Integration in den KB des Netzes benutzt werden.
	 * 
	 * @param objekte
	 *            Objekt-Map
	 * @param dateiname
	 *            Dateiname
	 */
	public void erzeugeASSListe(
			final Map<String, KonfigurationsObjekt> objekte,
			final String dateiname) {
		PrintStream ps;
		try {
			ps = new PrintStream(new File(dateiname));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			return;
		}
		for (KonfigurationsObjekt ko : objekte.values()) {
			if (ko instanceof SINASS) {
				ps.println("<element pid=\"" + ko.getPid() + "\"/>");
			}
		}

		ps.close();
	}

	/**
	 * Erzeugt eine KML-Datei des neu erzeugten Konfigurationsbereiches.
	 * 
	 * @param kd
	 *            Konfigurationsdaten
	 * @param kmlfile
	 *            KML-Datei
	 */
	private void erzeugeKML(KonfigurationsDaten kd, File kmlfile) {

		KML kml = new KML(kmlfile);

		kml.start("SIN", "importiertes Netz");

		kmlStyles(kml);

		String styleass = "ASS_STYLE";
		String styleiss = "ISS_STYLE";

		printTyp(kml, "de.bsvrz.kex.bwsin.konfig.kb.SINASS", styleass, "ASS",
				kd.getOjekteMap().values());
		printTyp(kml, "de.bsvrz.kex.bwsin.konfig.kb.SINISS", styleiss, "ISS",
				kd.getOjekteMap().values());

		kml.ende();
	}

	/**
	 * Erzeugt Zusatzinformationen zu den Teilsegmenten.
	 * 
	 * @param dateiname
	 *            Dateiname
	 * @param objekte
	 *            Konfigurationsobjekte
	 */
	private void erzeugeSTSInformationen(final String dateiname,
			final Collection<KonfigurationsObjekt> objekte) {
		PrintStream ps;
		List<String> bkliste = new ArrayList<String>();
		try {
			ps = new PrintStream(new File(dateiname));
		} catch (FileNotFoundException e) {
			System.err.println("Fehler beim Anlegen der Datei: " + dateiname);
			e.printStackTrace();
			return;
		}

		AsbTests at = new AsbTests();
		at.erzeugeInfo(asbzusatzinfodatei, objekte);

		for (KonfigurationsObjekt o : objekte) {
			if (o instanceof SINSTS) {
				SINSTS sts = (SINSTS) o;

				if (sts.getBetriebskilometer().size() > 1) {
					bkliste.add(o.getPid());
				}
			}
		}

		ps
				.println("=====================================================================================");
		ps
				.println("Strassenteilsegmente mit mehr als einer Betriebskilometerangabe:");
		ps
				.println("=====================================================================================");
		for (String s : bkliste) {
			ps.println(s);
		}

		ps.close();
	}

	/**
	 * Erzeugt Zusatzinformationen zum Import.
	 * 
	 * @param kd
	 *            KonfigurationsDaten
	 */
	private void erzeugeZusatzinformationen(final KonfigurationsDaten kd) {
		erzeugeASSListe(
				kd.getOjekteMap(),
				"/home/gysi/workspace/DAV-Neu/Projekte/VRZ3BaWue/konfiguration/kv.bw.bitctrl/xml/ass.txt");

		erzeugeKML(
				kd,
				new File(
						"/home/gysi/workspace/DAV-Neu/Projekte/VRZ3BaWue/konfiguration/kv.bw.bitctrl/xml/kb.sin.kml"));

		erzeugeSTSInformationen(zusatzinfodatei, kd.getOjekteMap().values());
	}

	public String getLocationCode(String location) {
		// potentiell mehrere Angaben
		// --> nehme die erste Location
		if (location.contains(";")) {
			String lc1 = location.substring(0, location.indexOf(';'));
			return lc1.substring(lc1.indexOf('.') + 1);
		} else {
			return location.substring(location.indexOf('.') + 1);
		}
	}

	public void initDav() {
		String[] commandArguments = {
				// "-datenverteiler=127.0.0.1:8083",
				// "-datenverteiler=10.44.44.10:8083",
				"-datenverteiler=192.168.1.137:8083",
				"-benutzer=Tester",
				"-authentifizierung=/home/gysi/workspace/DAV-Neu/Projekte/VRZ3BaWue/skripte-bash/passwd",
				// "-debugSetLoggerAndLevel=INFO",
				"-debugSetLoggerAndLevel=de.bsvrz.kex.bwsin:FINE",
				"-debugLevelFileText=FINE", "-debugFilePath=.",
				"-debugLevelStdErrText=FINE" };
		ArgumentList argumentList = new ArgumentList(commandArguments);
		try {
			String myname = "SIN-TEST";
			Debug.init(myname, argumentList);
			ClientDavParameters parameters = new ClientDavParameters(
					commandArguments);
			parameters.setApplicationName(myname);
			dav = new ClientDavConnection(parameters);
			dav.connect();
			dav.login();
		} catch (NoSuchElementException e) {
			// TODO Automatisch erstellter Catch-Block
			e.printStackTrace();
		} catch (MissingParameterException e) {
			// TODO Automatisch erstellter Catch-Block
			e.printStackTrace();
		} catch (CommunicationError e) {
			// TODO Automatisch erstellter Catch-Block
			e.printStackTrace();
		} catch (ConnectionException e) {
			// TODO Automatisch erstellter Catch-Block
			e.printStackTrace();
		} catch (InconsistentLoginException e) {
			// TODO Automatisch erstellter Catch-Block
			e.printStackTrace();
		}

	}

	public void initKB(ClientDavConnection dav) {
		kb = new SINKonfigurationsBereich();
		_kbfile = new File(
				"/home/gysi/workspace/DAV-Neu/Projekte/VRZ3BaWue//konfiguration/kv.bw.bitctrl/xml/"
						+ kbPid + ".xml");

		kb.initialisiere(dav, _kbfile.getAbsolutePath(), "kv.bw.bitctrl",
				kbPid, "SIN TEST", "gysi.kb.info");
	}

	public void initLcl() {
		// LCL _tlcl = new LCL();
		String _lclFile = "/home/gysi/Documents/BaW/LCL5.02.D.NW-060309.csv";
		try {
			LCL.initialisiere(_lclFile);
		} catch (IllegalArgumentException e) {
			// TODO Automatisch erstellter Catch-Block
			e.printStackTrace();
		} catch (FileNotFoundException e) {
			// TODO Automatisch erstellter Catch-Block
			e.printStackTrace();
		}
	}

	public void initOkstra() {
		Properties props = new Properties();

		// props.put("Pfad",
		// "/home/gysi/Documents/BaW/SIN/VRZ_BW_DSK_MIF_1_1/");

		Map<String, Object> m = new HashMap<String, Object>();

		// String URLstr =
		// "http://services.interactive-instruments.de/vrzbw/vrz3-wfs.exe";

		// String URLstr = "http://10.44.44.19/cgi-bin/vrz3-wfs.exe";

		String URLstr = "file:///home/gysi/workspace/de.bsvrz.kex.bwsin/okstrads_linux_neu";
		props.put("FILESTORE", Boolean.TRUE);

		URL url;
		try {
			url = new URL(URLstr);
		} catch (MalformedURLException e) {
			// TODO Automatisch erstellter Catch-Block
			e.printStackTrace();
			return;
		}
		props.put("SERVER_URL", url);

		// props.put(WFSDataStoreFactory.TIMEOUT.key,new Integer(3000000));
		// props.put(WFSDataStoreFactory.PROTOCOL.key, Boolean.FALSE); // false
		// == GET, true == POST
		// props.put(WFSDataStoreFactory.BUFFER_SIZE.key, new Integer(5000000));
		// // false == GET, true == POST

		try {
			odb = new Okstra(props);
		} catch (ImportException e) {
			// TODO Automatisch erstellter Catch-Block
			e.printStackTrace();
		}
	}

	/**
	 * Erzeugt die Style-Informationen der KML-Datei.
	 * 
	 * @param kml
	 *            KML
	 */
	private void kmlStyles(KML kml) {
		kml.appendLine("\t\t\t<Style id=\"" + "ASS_STYLE" + "\">");
		kml.appendLine("\t\t\t\t<LineStyle>");
		kml.appendLine("\t\t\t\t\t<color>ffff0000</color>");
		kml.appendLine("\t\t\t\t\t<width>4</width>");
		kml.appendLine("\t\t\t\t</LineStyle>");
		kml.appendLine("\t\t\t</Style>");

		kml.appendLine("\t\t\t<Style id=\"" + "ISS_STYLE" + "\">");
		kml.appendLine("\t\t\t\t<LineStyle>");
		kml.appendLine("\t\t\t\t\t<color>ff00ffff</color>");
		kml.appendLine("\t\t\t\t\t<width>4</width>");
		kml.appendLine("\t\t\t\t</LineStyle>");
		kml.appendLine("\t\t\t</Style>");
	}

	/**
	 * Gibt Koordinaten als KML aus.
	 * 
	 * @param kml
	 *            KML
	 * @param punkte
	 *            Koordinatenliste
	 */
	private void printKoordinaten(final KML kml,
			final ArrayList<WGS84Punkt> punkte) {
		kml.appendLine("<LineString>");
		kml.inclevel();
		kml.appendLine("<extrude>1</extrude>");
		kml.appendLine("<tessellate>1</tessellate>");
		kml.appendLine("<altitudeMode>absolute</altitudeMode>");
		kml.appendLine("<coordinates>");
		DecimalFormat df = new DecimalFormat(".#############");
		DecimalFormatSymbols dfs = new DecimalFormatSymbols();
		dfs.setDecimalSeparator('.');
		df.setDecimalFormatSymbols(dfs);

		for (WGS84Punkt punkt : punkte) {
			kml.appendLine(df.format(punkt.getLaenge()) + ","
					+ df.format(punkt.getBreite()));
		}

		kml.appendLine("</coordinates>");
		kml.declevel();
		kml.appendLine("</LineString>");
	}

	/**
	 * Erzeugt ein KML-Placemark Element.
	 * 
	 * @param kml
	 *            KML
	 * @param ko
	 *            KOnfigurationsobjekt
	 * @param style
	 *            Style
	 * @param punkte
	 *            Punktliste
	 */
	private void printPlacemark(final KML kml, final KonfigurationsObjekt ko,
			final String style, final ArrayList<WGS84Punkt> punkte) {
		kml.appendLine("<Placemark>");
		kml.inclevel();
		kml.appendLine("<name>" + ko.getPid() + "</name>");
		kml.appendLine("<description>" + ko.getName() + "</description>");
		kml.appendLine("<styleUrl>#" + style + "</styleUrl>");
		printKoordinaten(kml, punkte);
		kml.declevel();
		kml.appendLine("</Placemark>");
	}

	/**
	 * Ausgabe eines DaV-Typs als KML.
	 * 
	 * @param kml
	 *            KML
	 * @param clazz
	 *            Klassenname des Typs
	 * @param style
	 *            zu benutzender Style
	 * @param folder
	 *            KML-Ordner
	 * @param objekte
	 *            Objektliste
	 */
	private void printTyp(final KML kml, final String clazz,
			final String style, final String folder,
			final Collection<KonfigurationsObjekt> objekte) {

		kml.inclevel();
		kml.appendLine("<Folder><name>" + folder
				+ "</name><open>0</open><visibility>0</visibility>");

		for (KonfigurationsObjekt o : objekte) {
			if (o.getClass().getName().equals(clazz)) {
				ArrayList<WGS84Punkt> punkte = new ArrayList<WGS84Punkt>();
				SINStrassenSegment sss = (SINStrassenSegment) o;

				for (KonfigurationsObjekt kos : sss.getTeilSegmente()) {
					punkte.addAll(kos.getKoordinaten());
				}

				printPlacemark(kml, o, style, punkte);
			}
		}

		kml.appendLine("</Folder>");
		kml.declevel();
	}

	/**
	 * Testfunktion Import.
	 */
	@Test
	public void testImport() {

		initDav();
		initOkstra();
		initKB(dav);
		initLcl();

		SINAktuelleKonfiguration ak = new SINAktuelleKonfiguration();
		ak.initialisiere(dav, kbPid);

		KonfigurationsDaten kd = new KonfigurationsDaten(ak, dav, odb);

		// SINProperties.getInstance().setProperty(
		// SINProperties.PROP_KEY_XMLDEBUG, "ja");

		// SINProperties.getInstance().setProperty(SINProperties.PROP_KEY_TMCREFERENZEN,
		// "nein"); //$NON-NLS-1$

		try {
			kd.importiere();
			kd.exportKB(kb);

			kb.generiereXML();

			Abweichungsliste abl = new Abweichungsliste(ak, kd.getOjekteMap(),
					_kbfile);

			abl.erstellen();

			abl.print(System.out);
			abl
					.inDatei("/home/gysi/workspace/DAV-Neu/Projekte/VRZ3BaWue/konfiguration/kv.bw.bitctrl/xml/abweichungen.txt");
		} catch (ImportException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

		erzeugeZusatzinformationen(kd);
	}
}
