/*
 * Segment 2 (KEx),  SWE 2.Statischer Import Netz
 * Copyright (C) 2007 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weienfelser Strae 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.bwsin.konfig.kb;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import org.geotools.feature.Feature;

import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.kex.bwsin.konfig.AsbStationierung;
import de.bsvrz.kex.bwsin.konfig.BetriebsKilometer;
import de.bsvrz.kex.bwsin.konfig.BetriebsKilometerPunkt;
import de.bsvrz.kex.bwsin.konfig.ImportException;
import de.bsvrz.kex.bwsin.konfig.StrassenTeilAbschnitt;
import de.bsvrz.kex.bwsin.konfig.Verkehrsrichtung;
import de.bsvrz.kex.bwsin.konfig.okstra.IllegalAttributeConversionException;
import de.bsvrz.kex.bwsin.konfig.okstra.Okstra;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraAnzahlFahrstreifen;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraFNeigung;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraFeatures;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraMaxGeschwindigkeit;
import de.bsvrz.kex.bwsin.konfig.okstra.OkstraUeberholverbot;
import de.bsvrz.sys.funclib.bitctrl.geolib.WGS84Polygon;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.kbgen.ProxyConfigurationObject;
import de.bsvrz.sys.funclib.kbgen.ProxyDataDescription;

/**
 * SWE-interne Repr&auml;sentation eines Stra&szlig;enteilsegmentes.
 * 
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: SINSTS.java 20865 2009-12-22 15:06:01Z gieseler $
 * 
 */
public class SINSTS extends AbstraktKonfigurationsObjekt {

	/**
	 * Verkehrsbeschr&auml;nkung.
	 */
	private class Verkehrsbeschraenkung {

		/**
		 * Typ-Nummer im DAV.
		 */
		int davtypBeschraenkung = 0;

		/**
		 * Zeitbereich der Beschr&auml;nkung.
		 */
		Zeitbereich zeitbereich;

		/**
		 * Konstruktor.
		 * 
		 * @param typ
		 *            Typ-Nummer im DAV
		 * @param zeit
		 *            Zeitbereich
		 */
		Verkehrsbeschraenkung(final int typ, final Zeitbereich zeit) {
			davtypBeschraenkung = typ;
			zeitbereich = zeit;
			if (zeitbereich == null) {
				zeitbereich = new Zeitbereich();
			}
		}
	}

	/**
	 * DebugLogger fr Debug-Ausgaben.
	 */
	private static final Debug DEBUG = Debug.getLogger();

	/**
	 * DAV-Konstante ASB-Richtung unbekannt.
	 */
	public static final int DAV_ASB_RICHTUNG_UNBEKANNT = 0;

	/**
	 * DAV-Konstante ASB-Richtung In-Stationierungsrichtung.
	 */
	public static final int DAV_ASB_INSTATIONIERUNGSRICHTUNG = 1;

	/**
	 * DAV-Konstante ASB-Richtung Gegen-Stationierungsrichtung.
	 */
	public static final int DAV_ASB_GEGENSTATIONIERUNGSRICHTUNG = 2;

	/**
	 * Die Liste der ASB-Stationierungen.
	 */
	private List<AsbStationierung> asbStationierungen = new ArrayList<AsbStationierung>();

	/**
	 * Die Liste der Betriebskilometer-Angaben.
	 */
	private final List<BetriebsKilometerPunkt> betriebskilometer = new ArrayList<BetriebsKilometerPunkt>();

	/**
	 * Die Verkehrsrichtung.
	 */
	private Verkehrsrichtung verkehrsRichtung = Verkehrsrichtung.RICHTUNG_UNBEKANNT;

	/**
	 * Die Liste der Features, die Eigenschaften dieses STS bestimmen.
	 */
	protected List<Feature> eigenschaftenFeatures = new ArrayList<Feature>();

	/**
	 * Konstruktur mit PID und Name.
	 * 
	 * @param pid
	 *            PID
	 * @param name
	 *            Name
	 */
	public SINSTS(String pid, String name) {
		super(pid, name);
	}

	/**
	 * Konstruktur mit PID, Name und Teilabschnitt.
	 * 
	 * @param pid
	 *            PID
	 * @param name
	 *            Name
	 * @param teilabschnitt
	 *            zugeordneter Stra&szlig;enteilabschnitt
	 */
	public SINSTS(String pid, String name,
			final StrassenTeilAbschnitt teilabschnitt) {
		super(pid, name);
		setKoordinaten(teilabschnitt.getKoordinaten());
		setLaenge(teilabschnitt.laenge());
		// this.teilabschnitt = teilabschnitt;
		eigenschaftenFeatures = teilabschnitt.getEigenschaften();
		asbStationierungen = teilabschnitt.getAsbStationierungen();
	}

	/**
	 * Bestimmt den Zeitbereich einer .
	 * 
	 * @param feature
	 *            Feature Verkehrsbeschr&auml;nkung
	 * @param zeitattribut
	 *            Name des Attributes des Zeitbereiches.
	 * @return Zeitbereich
	 */
	private Zeitbereich bestimmeZeitbereichBeschraenkung(final Feature feature,
			final String zeitattribut) {
		Zeitbereich zeitbereich = null;
		String zeitbereichattr = (String) feature.getAttribute(zeitattribut);

		if ((zeitbereichattr != null) && zeitbereichattr.length() != 0) {
			try {
				zeitbereich = new Zeitbereich(Okstra
						.getFeatureReferenz(zeitbereichattr));
			} catch (ImportException e) {
				Debug.getLogger().error(e.getMessage());
			}
		}
		return zeitbereich;
	}

	/**
	 * <p>
	 * <b>DAV-Zust&auml;nde</b>
	 * </p>
	 * <p>
	 * </p>
	 * <table border="1" width="100%"> <tbody>
	 * <tr bgcolor="#cccccc">
	 * <th>Name</th>
	 * <th>Wert</th>
	 * <th>Info</th>
	 * </tr>
	 * <tr>
	 * <td>sonstige</td>
	 * <td align="right">0</td>
	 * <td>Nicht n&auml;her spezifizierte Einschr&auml;nkung.<br>
	 * </td>
	 * </tr>
	 * <tr>
	 * <td>&Uuml;berholverbot</td>
	 * <td align="right">1</td>
	 * <td>&Uuml;berholverbot allgemein.<br>
	 * </td>
	 * </tr>
	 * <tr>
	 * <td>&Uuml;berholverbot f&uuml;r LKW</td>
	 * <td align="right">2</td>
	 * <td>&Uuml;berholverbot f&uuml;r LKW.<br>
	 * </td>
	 * </tr>
	 * <tr>
	 * <td>&Uuml;berholverbot bei N&auml;sse</td>
	 * <td align="right">3</td>
	 * <td>&Uuml;berholverbot bei N&auml;sse<br>
	 * </td>
	 * </tr>
	 * <tr>
	 * <td>H&ouml;chstgewindigkeit Allgemein</td>
	 * <td align="right">4</td>
	 * <td>Allgemeine statische Geschwindigkeitsbeschr&auml;nkung.<br>
	 * </td>
	 * </tr>
	 * <tr>
	 * <td>H&ouml;chstgeschwindigkeit 10 km/h</td>
	 * <td align="right">13</td>
	 * <td>Allgemeine statische Geschwindigkeitsbeschr&auml;nkung auf 10 km/h.<br>
	 * </td>
	 * </tr>
	 * <tr>
	 * <td>H&ouml;chstgeschwindigkeit 20 km/h</td>
	 * <td align="right">14</td>
	 * <td>Allgemeine statische Geschwindigkeitsbeschr&auml;nkung auf 20 km/h.<br>
	 * </td>
	 * </tr>
	 * <tr>
	 * <td>H&ouml;chstgeschwindigkeit 30 km/h</td>
	 * <td align="right">15</td>
	 * <td>Allgemeine statische Geschwindigkeitsbeschr&auml;nkung auf 30 km/h.<br>
	 * </td>
	 * </tr>
	 * <tr>
	 * <td>H&ouml;chstgeschwindigkeit 40 km/h</td>
	 * <td align="right">5</td>
	 * <td>Allgemeine statische Geschwindigkeitsbeschr&auml;nkung auf 40 km/h.<br>
	 * </td>
	 * </tr>
	 * <tr>
	 * <td>H&ouml;chstgeschwindigkeit 50 km/h</td>
	 * <td align="right">6</td>
	 * <td>Allgemeine statische Geschwindigkeitsbeschr&auml;nkung auf 50 km/h.<br>
	 * </td>
	 * </tr>
	 * <tr>
	 * <td>H&ouml;chstgeschwindigkeit 60 km/h</td>
	 * <td align="right">7</td>
	 * <td>Allgemeine statische Geschwindigkeitsbeschr&auml;nkung auf 60 km/h.<br>
	 * </td>
	 * </tr>
	 * <tr>
	 * <td>H&ouml;chstgeschwindigkeit 70 km/h</td>
	 * <td align="right">8</td>
	 * <td>Allgemeine statische Geschwindigkeitsbeschr&auml;nkung auf 70 km/h.<br>
	 * </td>
	 * </tr>
	 * <tr>
	 * <td>H&ouml;chstgeschwindigkeit 80 km/h</td>
	 * <td align="right">9</td>
	 * <td>Allgemeine statische Geschwindigkeitsbeschr&auml;nkung auf 80 km/h.<br>
	 * </td>
	 * </tr>
	 * <tr>
	 * <td>H&ouml;chstgeschwindigkeit 100 km/h</td>
	 * <td align="right">10</td>
	 * <td>Allgemeine statische Geschwindigkeitsbeschr&auml;nkung auf 100 km/h.<br>
	 * </td>
	 * </tr>
	 * <tr>
	 * <td>H&ouml;chstgeschwindigkeit 120 km/h</td>
	 * <td align="right">11</td>
	 * <td>Allgemeine statische Geschwindigkeitsbeschr&auml;nkung auf 120 km/h.<br>
	 * </td>
	 * </tr>
	 * <tr>
	 * <td>H&ouml;chstgeschwindigkeit 130 km/h</td>
	 * <td align="right">12</td>
	 * <td>Allgemeine statische Geschwindigkeitsbeschr&auml;nkung auf 130 km/h.<br>
	 * </td>
	 * </tr>
	 * </tbody></table>
	 * 
	 * @return Liste von Beschr&auml;nkungen
	 */

	private List<Verkehrsbeschraenkung> bildeBeschraenkungen() {
		List<Verkehrsbeschraenkung> ret = new ArrayList<Verkehrsbeschraenkung>();

		for (Feature f : eigenschaftenFeatures) {
			if (OkstraFeatures.GESCHWINDIGKEIT
					.istFeatureTyp(f.getFeatureType())) {
				try {
					if ((OkstraMaxGeschwindigkeit.bestimmeVerkehrsRichtung(f
							.getID())) == verkehrsRichtung) {
						ret.add(bildeGeschwindigkeitsbeschraenkung(f));
					}
				} catch (ImportException e) {
					DEBUG.error(e.getMessage());
				}
			}

			if (OkstraFeatures.UEBERHOLVERBOT.istFeatureTyp(f.getFeatureType())) {
				try {
					if ((OkstraUeberholverbot.bestimmeVerkehrsRichtung(f
							.getID())) == verkehrsRichtung) {
						ret.add(bildeUeberholverbotbeschraenkung(f));
					}
				} catch (ImportException e) {
					DEBUG.error(e.getMessage());
				}
			}
		}

		return ret;
	}

	/**
	 * Erzeugt eine Geschwindigkeitsbeschr&auml;nkung.
	 * 
	 * @param f
	 *            Feature der Beschr&auml;nkung
	 * @return Verkehrsbeschr&auml;nkung
	 */
	private Verkehrsbeschraenkung bildeGeschwindigkeitsbeschraenkung(Feature f) {
		int maxKmh = 0;

		try {
			maxKmh = Okstra
					.getIntAttribute(f, OkstraMaxGeschwindigkeit.MAX_KMH);
		} catch (IllegalAttributeConversionException e) {
			DEBUG.error(e.getMessage());
			return null;
		}

		int davWert = 4;

		switch (maxKmh) {
		case 10:
			davWert = 13;
			break;
		case 20:
			davWert = 14;
			break;
		case 30:
			davWert = 15;
			break;
		case 40:
			davWert = 5;
			break;
		case 50:
			davWert = 6;
			break;
		case 60:
			davWert = 7;
			break;
		case 70:
			davWert = 8;
			break;
		case 80:
			davWert = 9;
			break;
		case 100:
			davWert = 10;
			break;
		case 120:
			davWert = 11;
			break;
		case 130:
			davWert = 12;
			break;
		default:
			davWert = 4;
		}

		// Grund (Zuordnung unklar)
		// String grund = (String)
		// f.getAttribute(OkstraMaxGeschwindigkeit.GRUND);

		Zeitbereich zeitbereich = bestimmeZeitbereichBeschraenkung(f,
				OkstraMaxGeschwindigkeit.ZEITRAUM);

		return new Verkehrsbeschraenkung(davWert, zeitbereich);
	}

	/**
	 * Erzeugt eine &Uuml;berholverbotbeschr&auml;nkung.
	 * 
	 * @param f
	 *            Feature der Beschr&auml;nkung
	 * @return Verkehrsbeschr&auml;nkung
	 */
	private Verkehrsbeschraenkung bildeUeberholverbotbeschraenkung(Feature f) {
		Object[] fzarten = Okstra.getInstance().getSpecialAttributeList(
				OkstraFeatures.UEBERHOLVERBOT, f.getID(),
				OkstraUeberholverbot.FAHRZEUGART);

		int davWert = 1;

		if (fzarten.length > 0) {
			for (int i = 0; i < fzarten.length; i++) {
				String fzart = (String) fzarten[i];
				if (fzart.equalsIgnoreCase("Lkw")) {
					davWert = 2;
					// Naesse ?
				}
			}
		}

		Zeitbereich zeitbereich = bestimmeZeitbereichBeschraenkung(f,
				OkstraUeberholverbot.ZEITRAUM);

		return new Verkehrsbeschraenkung(davWert, zeitbereich);
	}

	/**
	 * Bestimmt die Anzahl der Fahrstreifen.
	 * 
	 * @return Anzahl der Fahrstreifen
	 */
	private int getAnzahlFahrstreifen() {
		int anzahlfahrstreifen = 0; // unbekannt

		for (Feature f : eigenschaftenFeatures) {
			if (OkstraFeatures.ANZAHLFAHRSTREIFEN.istFeatureTyp(f
					.getFeatureType())) {
				try {
					if (verkehrsRichtung == Verkehrsrichtung.RICHTUNG_GEGEN_STATIONIERUNGSRICHTUNG) {
						anzahlfahrstreifen = Okstra
								.getIntAttribute(
										f,
										OkstraAnzahlFahrstreifen.FAHRSTREIFEN_GEGENRICHTUNG);
					} else {
						anzahlfahrstreifen = Okstra.getIntAttribute(f,
								OkstraAnzahlFahrstreifen.FAHRSTREIFEN_RICHTUNG);
					}
				} catch (IllegalAttributeConversionException e) {
					DEBUG.error(e.getMessage());
				}
			}
		}

		return (anzahlfahrstreifen);
	}

	/**
	 * Gibt die Liste der ASB-Stationierungen zur&uuml;ck.
	 * 
	 * @return Liste der ASB-Stationierungen
	 */
	public List<AsbStationierung> getAsbStationierungen() {
		return asbStationierungen;
	}

	/**
	 * Gibt die Liste der Betriebskilometerpunkte zur&uuml;ck.
	 * 
	 * @return Liste der Betriebskilometerpunkte
	 */
	public List<BetriebsKilometerPunkt> getBetriebskilometer() {
		return betriebskilometer;
	}

	/**
	 * Gibt den Wert des Attributes 'VerkehrsRichtung' des DAV f&uuml;r die
	 * Verkehrsrichtung zur&uuml;ck.
	 * 
	 * @param okstrarichtung
	 *            Richtung aus den Importdaten
	 * 
	 * @return ASB-Richtungswert entsprechend Datenkatalog
	 */
	private int getDavAsbRichtung(final Verkehrsrichtung okstrarichtung) {
		if (okstrarichtung == Verkehrsrichtung.RICHTUNG_IN_STATIONIERUNGSRICHTUNG) {
			return DAV_ASB_INSTATIONIERUNGSRICHTUNG;
		} else if (okstrarichtung == Verkehrsrichtung.RICHTUNG_GEGEN_STATIONIERUNGSRICHTUNG) {
			return DAV_ASB_GEGENSTATIONIERUNGSRICHTUNG;
		} else {
			return DAV_ASB_RICHTUNG_UNBEKANNT;
		}
	}

	/**
	 * Bestimmt die L&auml;ngsneigung.
	 * 
	 * @return L&auml;ngsneigung
	 */
	private double getLaengsneigung() {
		double steigung = -200; // unbekannt

		for (Feature f : eigenschaftenFeatures) {
			if (OkstraFeatures.FAHRBAHNNEIGUNG
					.istFeatureTyp(f.getFeatureType())) {
				try {
					if (OkstraFeatures.FAHRBAHNNEIGUNG.istFeatureTyp(f
							.getFeatureType())) {
						steigung = Okstra.getIntAttribute(f,
								OkstraFNeigung.STEIGUNG);
					}
				} catch (IllegalAttributeConversionException e) {
					DEBUG.error(e.getMessage());
				}
			}
		}

		return (steigung);
	}

	/**
	 * Versorgt alle Attribute und erzeugt ein {@link ProxyConfigurationObject}.
	 * 
	 * @param dav
	 *            Datenverteiler-Verbindung
	 * 
	 * @return {@link ProxyConfigurationObject}
	 */
	@SuppressWarnings("nls")
	public ProxyConfigurationObject getProxyObjekt(ClientDavConnection dav) {
		ProxyConfigurationObject pco = new ProxyConfigurationObject(pid, name,
				dav.getDataModel().getType("typ.straenTeilSegment"));

		davtyp = pco.getType();

		// typ.linie
		Data data = createData(dav, "atg.linie", "asp.eigenschaften");
		data.getScaledValue("Lnge").set(getLaenge());
		pco.addConfigurationData(new ProxyDataDescription(data, null));

		// typ.LinieXY
		data = createData(dav, "atg.linienKoordinaten", "asp.eigenschaften");
		data.setToDefault();
		Data.NumberArray x = data.getScaledArray("x");
		Data.NumberArray y = data.getScaledArray("y");

		x.setLength(koordinaten.size());
		y.setLength(koordinaten.size());

		for (int i = 0; i < koordinaten.size(); i++) {
			x.getValue(i).set(koordinaten.get(i).getLaenge());
			y.getValue(i).set(koordinaten.get(i).getBreite());
		}

		pco.addConfigurationData(new ProxyDataDescription(data, null));

		data = createData(dav, "atg.straenTeilSegment", "asp.eigenschaften");
		WGS84Polygon p = new WGS84Polygon(koordinaten);

		// Laenge aus Polygonzug
		data.getScaledValue("Lnge").set(p.laengeExakt());

		data.getScaledValue("AnzahlFahrStreifen").set(getAnzahlFahrstreifen());
		data.getScaledValue("SteigungGeflle").set(getLaengsneigung());
		pco.addConfigurationData(new ProxyDataDescription(data, null));

		List<Verkehrsbeschraenkung> beschraenkungen = bildeBeschraenkungen();
		data = createData(dav, "atg.straenTeilSegmentVerkehr",
				"asp.eigenschaften");
		data.setToDefault();
		Data.Array beschr = data.getItem("VerkehrsBeschrnkungen").asArray();
		beschr.setLength(beschraenkungen.size());
		for (int i = 0; i < beschraenkungen.size(); i++) {
			beschr.getItem(i).getItem("Typ").asUnscaledValue().set(
					beschraenkungen.get(i).davtypBeschraenkung);
			beschr.getItem(i).getItem("Zeitbereich").asTextValue().setText(
					beschraenkungen.get(i).zeitbereich.toString());
		}
		pco.addConfigurationData(new ProxyDataDescription(data, null));

		// Betriebskilometer aufsteigend nach Offset sortieren
		Collections.sort(betriebskilometer, new Comparator<BetriebsKilometerPunkt>() {
			public int compare(final BetriebsKilometerPunkt o1,
					final BetriebsKilometerPunkt o2) {
				return new Integer(o1.getOffset())
						.compareTo(new Integer(o2.getOffset()));
			}
		});
		data = createData(dav, "atg.betriebsKilometer", "asp.eigenschaften");
		Data.Array bk = data.getItem("BetriebsKilometer").asArray();
		bk.setLength(betriebskilometer.size());
		for (int i = 0; i < betriebskilometer.size(); i++) {
			bk.getItem(i).getItem("Offset").asUnscaledValue().set(
					betriebskilometer.get(i).getOffset());
			bk.getItem(i).getItem("Wert").asUnscaledValue().set(
					betriebskilometer.get(i).getDavWert());
			bk.getItem(i).getItem("BlockNummer").asTextValue().setText(
					"" + betriebskilometer.get(i).getBlocknummer());
		}
		pco.addConfigurationData(new ProxyDataDescription(data, null));

		// ASB-Stationierungen aufsteigend nach Offset sortieren
		Collections.sort(asbStationierungen, new Comparator<AsbStationierung>() {
			public int compare(final AsbStationierung o1,
					final AsbStationierung o2) {
				return new Integer(o1.getOffset())
						.compareTo(new Integer(o2.getOffset()));
			}
		});
		/*
		 * <attributDefinition pid="att.verkehrsRichtung"
		 * name="VerkehrsRichtung"> <info> <kurzinfo>Angabe des Richtungsbezugs
		 * der ASB-Stationierung relativ zur Richtung des Straensegments.</kurzinfo>
		 * </info> <ganzzahl bits="8"> <bereich skalierung="1" minimum="0"
		 * maximum="2"> </bereich> <zustand name="unbekannt" wert="0">
		 * </zustand> <zustand name="inStationierungsRichtung" wert="1">
		 * </zustand> <zustand name="gegenStationierungsRichtung" wert="2">
		 * </zustand> </ganzzahl> </attributDefinition>
		 */
		data = createData(dav, "atg.asbStationierung", "asp.eigenschaften");
		Data.Array asb = data.getItem("AsbStationierung").asArray();
		asb.setLength(asbStationierungen.size());
		for (int i = 0; i < asbStationierungen.size(); i++) {
			asb.getItem(i).getItem("Offset").asUnscaledValue().set(
					asbStationierungen.get(i).getOffset());
			asb.getItem(i).getItem("AnfangsKnoten").asTextValue().setText(
					asbStationierungen.get(i).getAnfangsKnoten());
			asb.getItem(i).getItem("EndKnoten").asTextValue().setText(
					asbStationierungen.get(i).getEndKnoten());
			asb.getItem(i).getItem("Anfang").asUnscaledValue().set(
					asbStationierungen.get(i).getStationierungAnfang());
			asb.getItem(i).getItem("Ende").asUnscaledValue().set(
					asbStationierungen.get(i).getStationierungEnde());
			asb.getItem(i).getItem("VerkehrsRichtung").asUnscaledValue().set(
					getDavAsbRichtung(asbStationierungen.get(i)
							.getVerkehrsRichtung()));
		}
		pco.addConfigurationData(new ProxyDataDescription(data, null));

		return pco;
	}

	/**
	 * Setzt die Verkehrsrichtung.
	 * 
	 * @param richtung
	 *            Verkehrsrichtung
	 */
	public void setRichtung(Verkehrsrichtung richtung) {
		this.verkehrsRichtung = richtung;
	}

	/**
	 * Ordnet die Betriebskilometer zu.
	 * 
	 * @param betriebsKilometer
	 *            Betriebskilometer
	 */
	public void setzteBetriebsKilometer(
			final BetriebsKilometer betriebsKilometer) {
		for (AsbStationierung asb : asbStationierungen) {
			List<BetriebsKilometerPunkt> bkpl = betriebsKilometer.getAoaPunkte(
					asb.getAOAId(), asb.getStationierungAnfang(), asb
							.getStationierungEnde());

			// Offset korrigieren
			for (BetriebsKilometerPunkt bkp : bkpl) {
				BetriebsKilometerPunkt bkpc = bkp.clone();
				if(asb.getVerkehrsRichtung() == Verkehrsrichtung.RICHTUNG_IN_STATIONIERUNGSRICHTUNG) {
					bkpc.setOffset(bkp.getStationierung()
						- asb.getStationierungAnfang() + asb.getOffset());
				}
				else if(asb.getVerkehrsRichtung() == Verkehrsrichtung.RICHTUNG_GEGEN_STATIONIERUNGSRICHTUNG) {
					bkpc.setOffset(asb.getStationierungEnde()- bkp.getStationierung()
							+ asb.getOffset());
				}
				else {
					DEBUG.error("Betriebskilometer kann nicht gesetzt werden, da Verkehrsrichtung unbekannt!");
					return;
				}
				betriebskilometer.add(bkpc);
			}
		}
	}
}
