package org.geotools.data.wfs;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLEncoder;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.GZIPInputStream;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.geotools.data.DataUtilities;
import org.geotools.data.FeatureReader;
import org.geotools.data.Query;
import org.geotools.data.Transaction;
import org.geotools.data.ows.FeatureSetDescription;
import org.geotools.data.ows.WFSCapabilities;
import org.geotools.feature.FeatureType;
import org.geotools.feature.SchemaException;
import org.geotools.referencing.CRS;
import org.geotools.xml.DocumentFactory;
import org.geotools.xml.gml.BWSIN_GMLComplexTypes;
import org.geotools.xml.gml.BWSIN_WFSFeatureTypeTransformer;
import org.geotools.xml.schema.Element;
import org.geotools.xml.schema.Schema;
import org.opengis.referencing.FactoryException;
import org.opengis.referencing.crs.CoordinateReferenceSystem;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

public class BWSIN_WFSDataStore extends WFSDataStore {

	static FeatureType BWSIN_parseDescribeFeatureTypeResponse(String typeName,
			Schema schema) throws SAXException {
		Element[] elements = schema.getElements();

		if (elements == null) {
			return null; // not found
		}

		Element element = null;

		String ttname = typeName.substring(typeName.indexOf(":") + 1);

		for (int i = 0; (i < elements.length) && (element == null); i++) {
			// HACK -- namspace related -- should be checking ns as opposed to
			// removing prefix
			if (typeName.equals(elements[i].getName())
					|| ttname.equals(elements[i].getName())) {
				element = elements[i];
			}
		}

		if (element == null) {
			return null;
		}

		// FeatureType ft = GMLComplexTypes.createFeatureType(element);

		// GYSI ????
		// ElementGrouping child =
		// ((org.geotools.xml.schema.ComplexType)element.getType()).getChild();
		// //AttributeType[] attrs =
		// (AttributeType[])GMLComplexTypes.getAttributes(element.getName(),child).toArray(new
		// AttributeType[]{,});
		//

		FeatureType ft = BWSIN_GMLComplexTypes.createFeatureType(element);

		return ft;
	}

	// Hilfsmethode, WFSDataStore zuerst einen Fehler erzeugt, wenn die URL
	// nicht correct ist
	public static URL createGetCapabilitiesRequest(URL host) {
		if (host == null) {
			return null;
		}

		String url = host.toString();

		if (host.getQuery() == null) {
			url += "?SERVICE=WFS&VERSION=1.0.0&REQUEST=GetCapabilities";
		} else {
			String t = host.getQuery().toUpperCase();

			if (t.indexOf("SERVICE") == -1) {
				url += "&SERVICE=WFS";
			}

			if (t.indexOf("VERSION") == -1) {
				url += "&VERSION=1.0.0";
			}

			if (t.indexOf("REQUEST") == -1) {
				url += "&REQUEST=GetCapabilities";
			}
		}

		try {
			return new URL(url);
		} catch (MalformedURLException e) {
			WFSDataStoreFactory.logger.warning(e.toString());

			return host;
		}
	}

	private final Map featureTypeCache = new HashMap();
	private final boolean tryGZIP = true;
	private int bufferSize = 10;

	private int timeout = 10000;

	private boolean lenient;

	protected BWSIN_WFSDataStore(URL arg0, Boolean arg1, String arg2,
			String arg3, int arg4, int arg5, boolean arg6) throws SAXException,
			IOException {
		super(arg0, arg1, arg2, arg3, arg4, arg5, arg6);
		// TODO Automatisch erstellter Konstruktoren-Stub
		WFSDataStoreFactory.logger.setLevel(Level.WARNING);
	}

	public BWSIN_WFSDataStore(URL host, Boolean protocol, String user,
			String pass, int timeout, int buffer, boolean tryGZIP2,
			boolean lenient) throws SAXException, IOException {
		super(host, protocol, user, pass, timeout, buffer, tryGZIP2, lenient);
		this.timeout = timeout;
		this.bufferSize = buffer;
		this.lenient = lenient;
		WFSDataStoreFactory.logger.setLevel(Level.WARNING);
	}

	/**
	 * If the field useGZIP is true Adds gzip to the connection accept-encoding
	 * property and creates a gzip inputstream (if server supports it).
	 * Otherwise returns a normal buffered input stream.
	 * 
	 * @param hc
	 *            the connection to use to create the stream
	 * @return an input steam from the provided connection
	 */
	InputStream BWSIN_getInputStream(HttpURLConnection hc) throws IOException {
		InputStream is = hc.getInputStream();

		if (tryGZIP) {
			if (hc.getContentEncoding() != null
					&& hc.getContentEncoding().indexOf("gzip") != -1) {
				is = new GZIPInputStream(is);
			}
		}
		is = new BufferedInputStream(is);
		// if( WFSDataStoreFactory.logger.isLoggable(Level.FINE) ){
		// is=new LogInputStream(is, WFSDataStoreFactory.logger, Level.FINE);
		// }
		// special logger for communication information only.
		if (Logger.getLogger("org.geotools.data.communication").isLoggable(
				Level.FINE)) {
			is = new LogInputStream(is, Logger
					.getLogger("org.geotools.data.communication"), Level.FINE);
		}
		return is;
	}

	private void findCapabilities(URL host) throws SAXException, IOException {

		// TODO support using POST for getCapabilities

		Object t = null;
		Map hints = new HashMap();
		hints.put(DocumentFactory.VALIDATION_HINT, Boolean.FALSE);
		try {
			HttpURLConnection hc = getConnection(host, auth, false);
			InputStream is = getInputStream(hc);
			t = DocumentFactory.getInstance(is, hints,
					WFSDataStoreFactory.logger.getLevel());
		} catch (Throwable e) {
			HttpURLConnection hc = getConnection(
					createGetCapabilitiesRequest(host), auth, false);
			InputStream is = getInputStream(hc);
			t = DocumentFactory.getInstance(is, hints,
					WFSDataStoreFactory.logger.getLevel());
		}
		if (t instanceof WFSCapabilities) {
			capabilities = (WFSCapabilities) t;
		} else {
			throw new SAXException(
					"The specified URL Should have returned a 'WFSCapabilities' object. Returned a "
							+ ((t == null) ? "null value." : (t.getClass()
									.getName() + " instance.")));
		}
	}

	private URL getDescribeFeatureTypeURLGet(String typeName)
			throws MalformedURLException {
		URL getUrl = capabilities.getDescribeFeatureType().getGet();
		Logger.getLogger("org.geotools.data.communication").fine(
				"Output: " + getUrl);

		if (getUrl == null) {
			return null;
		}

		String query = getUrl.getQuery();
		query = query == null ? null : query.toUpperCase();
		String url = getUrl.toString();

		if ((query == null) || "".equals(query)) {
			if ((url == null) || !url.endsWith("?")) {
				url += "?";
			}

			url += "SERVICE=WFS";
		} else {
			if (query.indexOf("SERVICE=WFS") == -1) {
				url += "&SERVICE=WFS";
			}
		}

		if ((query == null) || (query.indexOf("VERSION") == -1)) {
			url += "&VERSION=1.0.0";
		}

		if ((query == null) || (query.indexOf("REQUEST") == -1)) {
			url += "&REQUEST=DescribeFeatureType";
		}

		url += ("&TYPENAME=" + typeName);

		getUrl = new URL(url);
		return getUrl;
	}

	/**
	 * Constructs an AttributeType from a row in a ResultSet. The ResultSet
	 * contains the information retrieved by a call to getColumns() on the
	 * DatabaseMetaData object. This information can be used to construct an
	 * Attribute Type.
	 * 
	 * <p>
	 * This implementation construct an AttributeType using the default JDBC
	 * type mappings defined in JDBCDataStore. These type mappings only handle
	 * native Java classes and SQL standard column types. If a geometry type is
	 * found then getGeometryAttribute is called.
	 * </p>
	 * 
	 * <p>
	 * Note: Overriding methods must never move the current row pointer in the
	 * result set.
	 * </p>
	 * 
	 * @param metadataRs
	 *            The ResultSet containing the result of a
	 *            DatabaseMetaData.getColumns call.
	 * 
	 * @return The AttributeType built from the ResultSet.
	 * 
	 * @throws IOException
	 *             If an error occurs processing the ResultSet.
	 * @throws ParserConfigurationException
	 */
	/*
	 * protected AttributeType buildAttributeType(ResultSet metadataRs) throws
	 * IOException { try { final int TABLE_NAME = 3; final int COLUMN_NAME = 4;
	 * final int TYPE_NAME = 6; final int NULLABLE = 11; String typeName =
	 * metadataRs.getString(TYPE_NAME);
	 * 
	 * if (typeName.equals("geometry")) { String tableName =
	 * metadataRs.getString(TABLE_NAME); String columnName =
	 * metadataRs.getString(COLUMN_NAME); // check for nullability int nullCode =
	 * metadataRs.getInt( NULLABLE ); boolean nillable = true; switch( nullCode ) {
	 * case DatabaseMetaData.columnNoNulls: nillable = false; break;
	 * 
	 * case DatabaseMetaData.columnNullable: nillable = true; break;
	 * 
	 * case DatabaseMetaData.columnNullableUnknown: nillable = true; break; }
	 * 
	 * return getGeometryAttribute(tableName, columnName, nillable); } else {
	 * return buildAttributeType(metadataRs); } } catch (SQLException e) { throw
	 * new IOException("Sql error occurred: " + e.getMessage()); } }
	 */

	public Document getDocument(Query request)
			throws UnsupportedEncodingException, IOException, SAXException,
			ParserConfigurationException {
		URL getUrl = capabilities.getGetFeature().getGet();

		if (getUrl == null) {
			return null;
		}

		String query = getUrl.getQuery();
		query = query == null ? null : query.toUpperCase();
		String url = getUrl.toString();

		if ((query == null) || "".equals(query)) {
			if ((url == null) || !url.endsWith("?")) {
				url += "?";
			}

			url += "SERVICE=WFS";
		} else {
			if (query.indexOf("SERVICE=WFS") == -1) {
				url += "&SERVICE=WFS";
			}
		}

		if ((query == null) || (query.indexOf("VERSION") == -1)) {
			url += "&VERSION=1.0.0";
		}

		if ((query == null) || (query.indexOf("REQUEST") == -1)) {
			url += "&REQUEST=GetFeature";
		}

		if (request != null) {
			if (request.getMaxFeatures() != Query.DEFAULT_MAX) {
				url += ("&MAXFEATURES=" + request.getMaxFeatures());
			}
			// GYSI
			// if (request.getFilter() != null) {
			// if ( Filters.getFilterType(request.getFilter()) ==
			// FilterType.GEOMETRY_BBOX) {
			// String bb = printBBoxGet(((GeometryFilter)
			// request.getFilter()),request.getTypeName());
			// if(bb!=null)
			// url += ("&BBOX=" + URLEncoder.encode(bb, "UTF-8"));
			// } else {
			// if (Filters.getFilterType(request.getFilter()) == FilterType.FID)
			// {
			// FidFilter ff = (FidFilter) request.getFilter();
			//
			// if ((ff.getFids() != null) && (ff.getFids().length > 0)) {
			// url += ("&FEATUREID=" + ff.getFids()[0]);
			//
			// for (int i = 1; i < ff.getFids().length; i++) {
			// url += ("," + ff.getFids()[i]);
			// }
			// }
			// } else {
			// // rest
			// if (request.getFilter() != Filter.INCLUDE && request.getFilter()
			// != Filter.EXCLUDE) {
			// url += "&FILTER=" + URLEncoder.encode(
			// printFilter(request.getFilter()), "UTF-8");
			// }
			// }
			// }
			// }
		}

		url += ("&TYPENAME=" + URLEncoder
				.encode(request.getTypeName(), "UTF-8"));

		Logger.getLogger("org.geotools.data.wfs").fine(url);
		Logger.getLogger("org.geotools.data.communication").fine(
				"Output: " + url);
		getUrl = new URL(url);

		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder parser = factory.newDocumentBuilder();

		HttpURLConnection hc = getConnection(getUrl, auth, false);

		InputStream is = getInputStream(hc);

		Document doc = parser.parse(is);

		return doc;
	}

	/**
	 * @see org.geotools.data.DataStore#getFeatureReader(org.geotools.data.Query,
	 *      org.geotools.data.Transaction)
	 */
	@Override
	public FeatureReader getFeatureReader(Query query, Transaction transaction)
			throws IOException {
		// return strategy.getFeatureReader(query, transaction);
		try {
			return getFeatureReaderGet(query, transaction);
		} catch (SAXException e) {
			// TODO Automatisch erstellter Catch-Block
			e.printStackTrace();
			return null;
		}
	}

	// protected for testing
	@Override
	protected BWSIN_WFSFeatureReader getFeatureReaderGet(Query request,
			Transaction transaction) throws UnsupportedEncodingException,
			IOException, SAXException {
		URL getUrl = capabilities.getGetFeature().getGet();

		if (getUrl == null) {
			return null;
		}

		String query = getUrl.getQuery();
		query = query == null ? null : query.toUpperCase();
		String url = getUrl.toString();

		if ((query == null) || "".equals(query)) {
			if ((url == null) || !url.endsWith("?")) {
				url += "?";
			}

			url += "SERVICE=WFS";
		} else {
			if (query.indexOf("SERVICE=WFS") == -1) {
				url += "&SERVICE=WFS";
			}
		}

		if ((query == null) || (query.indexOf("VERSION") == -1)) {
			url += "&VERSION=1.0.0";
		}

		if ((query == null) || (query.indexOf("REQUEST") == -1)) {
			url += "&REQUEST=GetFeature";
		}

		if (request != null) {
			if (request.getMaxFeatures() != Query.DEFAULT_MAX) {
				url += ("&MAXFEATURES=" + request.getMaxFeatures());
			}
			// GYSI
			// if (request.getFilter() != null) {
			// if ( Filters.getFilterType(request.getFilter()) ==
			// FilterType.GEOMETRY_BBOX) {
			// String bb = printBBoxGet(((GeometryFilter)
			// request.getFilter()),request.getTypeName());
			// if(bb!=null)
			// url += ("&BBOX=" + URLEncoder.encode(bb, "UTF-8"));
			// } else {
			// if (Filters.getFilterType(request.getFilter()) == FilterType.FID)
			// {
			// FidFilter ff = (FidFilter) request.getFilter();
			//
			// if ((ff.getFids() != null) && (ff.getFids().length > 0)) {
			// url += ("&FEATUREID=" + ff.getFids()[0]);
			//
			// for (int i = 1; i < ff.getFids().length; i++) {
			// url += ("," + ff.getFids()[i]);
			// }
			// }
			// } else {
			// // rest
			// if (request.getFilter() != Filter.INCLUDE && request.getFilter()
			// != Filter.EXCLUDE) {
			// url += "&FILTER=" + URLEncoder.encode(
			// printFilter(request.getFilter()), "UTF-8");
			// }
			// }
			// }
			// }
		}

		url += ("&TYPENAME=" + URLEncoder
				.encode(request.getTypeName(), "UTF-8"));

		Logger.getLogger("org.geotools.data.wfs").fine(url);
		Logger.getLogger("org.geotools.data.communication").fine(
				"Output: " + url);
		getUrl = new URL(url);
		HttpURLConnection hc = getConnection(getUrl, auth, false);

		InputStream is = getInputStream(hc);
		WFSTransactionState ts = null;

		if (!(transaction == Transaction.AUTO_COMMIT)) {
			ts = (WFSTransactionState) transaction.getState(this);

			if (ts == null) {
				ts = new WFSTransactionState(this);
				transaction.putState(this, ts);
			}
		}

		WFSFeatureType schema = (WFSFeatureType) getSchema(request
				.getTypeName());

		FeatureType featureType;
		try {
			featureType = DataUtilities.createSubType(schema.delegate, request
					.getPropertyNames(), request.getCoordinateSystem());
		} catch (SchemaException e) {
			featureType = schema.delegate;
		}

		BWSIN_WFSFeatureReader ft = BWSIN_WFSFeatureReader.getFeatureReader(is,
				bufferSize, timeout, ts, new WFSFeatureType(schema.delegate,
						schema.getSchemaURI(), lenient));
		// WFSFeatureReader ft = WFSFeatureReader.getFeatureReader(is,
		// bufferSize,
		// timeout, ts, new WFSFeatureType(schema.delegate,
		// schema.getSchemaURI(), lenient));

		if (!featureType.equals(ft.getFeatureType())) {
			LOGGER
					.fine("Recasting feature type to subtype by using a ReTypeFeatureReader");
			// GYSI
			// return new ReTypeFeatureReader(ft, featureType, false);
			return null;
		} else {
			return ft;
		}

	}

	/**
	 * DOCUMENT ME!
	 * 
	 * @param typeName
	 *            DOCUMENT ME!
	 * 
	 * @return DOCUMENT ME!
	 * 
	 * @throws IOException
	 * 
	 * @see org.geotools.data.AbstractDataStore#getSchema(java.lang.String)
	 */
	@Override
	public FeatureType getSchema(String typeName) throws IOException {
		if (featureTypeCache.containsKey(typeName)) {
			return (FeatureType) featureTypeCache.get(typeName);
		}

		// TODO sanity check for request with capabilities obj

		FeatureType t = null;
		SAXException sax = null;
		IOException io = null;
		if (((protocol & POST_PROTOCOL) == POST_PROTOCOL) && (t == null)) {
			try {
				t = getSchemaPost(typeName);
			} catch (SAXException e) {
				// WFSDataStoreFactory.logger.warning(e.toString());
				sax = e;
			} catch (IOException e) {
				// WFSDataStoreFactory.logger.warning(e.toString());
				io = e;
			}
		}

		if (((protocol & GET_PROTOCOL) == GET_PROTOCOL) && (t == null)) {
			try {
				t = getSchemaGet(typeName);
			} catch (SAXException e) {
				// WFSDataStoreFactory.logger.warning(e.toString());
				sax = e;
			} catch (IOException e) {
				// WFSDataStoreFactory.logger.warning(e.toString());
				io = e;
			}
		}

		if (t == null && sax != null) {
			throw new IOException(sax.toString());
		}

		if (t == null && io != null) {
			throw io;
		}

		// set crs?
		FeatureSetDescription fsd = WFSCapabilities.getFeatureSetDescription(
				capabilities, typeName);
		String crsName = null;
		String ftName = null;
		if (fsd != null) {
			crsName = fsd.getSRS();
			ftName = fsd.getName();

			CoordinateReferenceSystem crs;
			try {
				if (crsName != null) {
					crs = CRS.decode(crsName);
					// GYSI
					t = BWSIN_WFSFeatureTypeTransformer.transform(t, crs);
				}
			} catch (FactoryException e) {
				WFSDataStoreFactory.logger.warning(e.getMessage());
			} catch (SchemaException e) {
				WFSDataStoreFactory.logger.warning(e.getMessage());
			}
		}

		// GYSI
		// if(ftName!=null){
		// try {
		// t =
		// FeatureTypeBuilder.newFeatureType(t.getAttributeTypes(),ftName==null?typeName:ftName,t.getNamespace(),t.isAbstract(),t.getAncestors(),t.getDefaultGeometry());
		//                
		// } catch (SchemaException e1) {
		// // WFSDataStoreFactory.logger.warning(e1.getMessage());
		// }
		// }
		try {
			URL url = getDescribeFeatureTypeURLGet(typeName);
			if (url != null) {
				t = new WFSFeatureType(t, new URI(url.toString()));
			}
		} catch (URISyntaxException e) {
			throw new RuntimeException(e);
		}
		if (t != null) {
			featureTypeCache.put(typeName, t);
		}

		return t;
	}

	// protected for testing
	@Override
	protected FeatureType getSchemaGet(String typeName) throws SAXException,
			IOException {
		URL getUrl = getDescribeFeatureTypeURLGet(typeName);
		Logger.getLogger("org.geotools.data.communication").fine(
				"Output: " + getUrl);
		if (getUrl == null) {
			return null;
		}
		HttpURLConnection hc = getConnection(getUrl, auth, false);

		// InputStream is = BWSIN_getInputStream(hc);
		InputStream is = getInputStream(hc);
		Schema schema = null;

		try {
			// schema = BWSIN_SchemaFactory.getInstance(null, is);
			// schema = SchemaFactory.getInstance(null, is);
			// schema = BWSIN_SchemaFactory.getInstance(new URI
			// ("http://schema.okstra.de/1009/okstra"), is);
			schema = BWSIN_SchemaFactory.getInstance(null, is);

			// URI ns = null;

			// try {
			// ns = new URI("http://schema.okstra.de/1009/okstra");
			// } catch (URISyntaxException e) {
			// // TODO Automatisch erstellter Catch-Block
			// e.printStackTrace();
			// }
			// schema = SINSchemaFactory.getInstance(ns, is);

			// } catch (URISyntaxException e) {
			// // TODO Automatisch erstellter Catch-Block
			// e.printStackTrace();
		} finally {
			is.close();
		}
		return BWSIN_parseDescribeFeatureTypeResponse(typeName, schema);
		// return parseDescribeFeatureTypeResponse(typeName, schema);
	}

	public InputSource getSource(Query request)
			throws UnsupportedEncodingException, IOException, SAXException,
			ParserConfigurationException {
		URL getUrl = capabilities.getGetFeature().getGet();

		if (getUrl == null) {
			return null;
		}

		String query = getUrl.getQuery();
		query = query == null ? null : query.toUpperCase();
		String url = getUrl.toString();

		if ((query == null) || "".equals(query)) {
			if ((url == null) || !url.endsWith("?")) {
				url += "?";
			}

			url += "SERVICE=WFS";
		} else {
			if (query.indexOf("SERVICE=WFS") == -1) {
				url += "&SERVICE=WFS";
			}
		}

		if ((query == null) || (query.indexOf("VERSION") == -1)) {
			url += "&VERSION=1.0.0";
		}

		if ((query == null) || (query.indexOf("REQUEST") == -1)) {
			url += "&REQUEST=GetFeature";
		}

		if (request != null) {
			if (request.getMaxFeatures() != Query.DEFAULT_MAX) {
				url += ("&MAXFEATURES=" + request.getMaxFeatures());
			}
			// GYSI
			// if (request.getFilter() != null) {
			// if ( Filters.getFilterType(request.getFilter()) ==
			// FilterType.GEOMETRY_BBOX) {
			// String bb = printBBoxGet(((GeometryFilter)
			// request.getFilter()),request.getTypeName());
			// if(bb!=null)
			// url += ("&BBOX=" + URLEncoder.encode(bb, "UTF-8"));
			// } else {
			// if (Filters.getFilterType(request.getFilter()) == FilterType.FID)
			// {
			// FidFilter ff = (FidFilter) request.getFilter();
			//
			// if ((ff.getFids() != null) && (ff.getFids().length > 0)) {
			// url += ("&FEATUREID=" + ff.getFids()[0]);
			//
			// for (int i = 1; i < ff.getFids().length; i++) {
			// url += ("," + ff.getFids()[i]);
			// }
			// }
			// } else {
			// // rest
			// if (request.getFilter() != Filter.INCLUDE && request.getFilter()
			// != Filter.EXCLUDE) {
			// url += "&FILTER=" + URLEncoder.encode(
			// printFilter(request.getFilter()), "UTF-8");
			// }
			// }
			// }
			// }
		}

		url += ("&TYPENAME=" + URLEncoder
				.encode(request.getTypeName(), "UTF-8"));

		Logger.getLogger("org.geotools.data.wfs").fine(url);
		Logger.getLogger("org.geotools.data.communication").fine(
				"Output: " + url);
		getUrl = new URL(url);

		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder parser = factory.newDocumentBuilder();

		HttpURLConnection hc = getConnection(getUrl, auth, false);

		InputStream is = getInputStream(hc);

		return new InputSource(is);
	}
}
